from unittest import (
    TestCase,
    mock,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.constraint.location import command as location_command
from pcs.common import (
    const,
    reports,
)

from pcs_test.tools.constraints_dto import get_all_constraints
from pcs_test.tools.custom_mock import RuleInEffectEvalMock
from pcs_test.tools.misc import dict_to_modifiers

RULE_ARGV_DEPRECATED = (
    "Specifying a rule as multiple arguments is deprecated and might be removed "
    "in a future release, specify the rule as a single string instead"
)


class TestRemoveLocationConstraint(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["constraint", "cib"])
        self.cib = mock.Mock(spec_set=["remove_elements"])
        self.constraint = mock.Mock(spec_set=["get_config"])
        self.constraint.get_config.return_value = get_all_constraints(
            RuleInEffectEvalMock({})
        )
        self.lib.cib = self.cib
        self.lib.constraint = self.constraint
        self.patch_warn = mock.patch(
            "pcs.cli.constraint.location.command.deprecation_warning"
        )
        self.mock_warn = self.patch_warn.start()

    def tearDown(self):
        self.patch_warn.stop()

    def _call_cmd(self, argv, modifiers=None):
        location_command.remove(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )
        self.mock_warn.assert_called_once_with(
            "This command is deprecated and will be removed. "
            "Please use 'pcs constraint delete' or 'pcs constraint remove' "
            "instead."
        )

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.constraint.get_config.assert_not_called()
        self.cib.remove_elements.assert_not_called()

    def test_duplicate_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["id1", "id2", "id1", "id2", "id3"])
        self.assertEqual(
            cm.exception.message, "duplicate arguments: 'id1', 'id2'"
        )
        self.constraint.get_config.assert_not_called()
        self.cib.remove_elements.assert_not_called()

    def test_unsupported_modifier(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["id1", "id2", "id1", "id2"], {"force": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--force' is not supported in this command",
        )
        self.constraint.get_config.assert_not_called()
        self.cib.remove_elements.assert_not_called()

    def test_location_constraints_ids_not_found(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(
                [
                    "id1",
                    "order-R7-G2-mandatory",
                    "location-R7-localhost-INFINITY",
                ]
            )
        self.assertEqual(
            cm.exception.message,
            (
                "Unable to find location constraints: 'id1', "
                "'order-R7-G2-mandatory'"
            ),
        )
        self.constraint.get_config.assert_called_once_with(evaluate_rules=False)
        self.cib.remove_elements.assert_not_called()

    def test_location_constraint_ids_found(self):
        location_constraint_ids = [
            "location-R7-non-existing-node--10000",
            "location-R7-another-one--INFINITY",
        ]
        self._call_cmd(location_constraint_ids)
        self.constraint.get_config.assert_called_once_with(evaluate_rules=False)
        self.cib.remove_elements.assert_called_once_with(
            location_constraint_ids
        )


class CreateWithRule(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["constraint_location", "env"])
        self.lib_module = mock.Mock(spec_set=["create_plain_with_rule"])
        self.lib.constraint_location = self.lib_module
        env = mock.Mock(spec_set=["report_processor"])
        self.lib.env = env
        self.report_processor = mock.Mock(
            spec_set=["set_report_item_preprocessor"]
        )
        self.lib.env.report_processor = self.report_processor

    def _call_cmd(self, argv, modifiers=None):
        location_command.create_with_rule(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.lib_module.create_plain_with_rule.assert_not_called()
        self.report_processor.set_report_item_preprocessor.assert_not_called()

    def test_not_enough_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd("R1 #uname eq node1".split())
        self.assertIsNone(cm.exception.message)
        self.lib_module.create_plain_with_rule.assert_not_called()
        self.report_processor.set_report_item_preprocessor.assert_not_called()

    def test_missing_rule_keyword(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd("R1 score=123 #uname eq node1".split())
        self.assertIsNone(cm.exception.message)
        self.lib_module.create_plain_with_rule.assert_not_called()
        self.report_processor.set_report_item_preprocessor.assert_not_called()

    def test_minimal(self):
        self._call_cmd(["R1", "rule", "#uname eq node1"])
        self.lib_module.create_plain_with_rule.assert_called_once_with(
            const.RESOURCE_ID_TYPE_PLAIN,
            "R1",
            "#uname eq node1",
            {},
            {},
            set(),
        )
        self.report_processor.set_report_item_preprocessor.assert_called_once()

    @mock.patch("pcs.cli.common.parse_args.deprecation_warning")
    def test_minimal_deprecated_form(self, mock_dw):
        self._call_cmd("R1 rule #uname eq node1".split())
        self.lib_module.create_plain_with_rule.assert_called_once_with(
            const.RESOURCE_ID_TYPE_PLAIN,
            "R1",
            "#uname eq node1",
            {},
            {},
            set(),
        )
        self.report_processor.set_report_item_preprocessor.assert_called_once()
        mock_dw.assert_called_once_with(RULE_ARGV_DEPRECATED)

    def test_resource_id(self):
        self._call_cmd(["resource%R1", "rule", "#uname eq node1"])
        self.lib_module.create_plain_with_rule.assert_called_once_with(
            const.RESOURCE_ID_TYPE_PLAIN,
            "R1",
            "#uname eq node1",
            {},
            {},
            set(),
        )
        self.report_processor.set_report_item_preprocessor.assert_called_once()

    def test_resource_pattern(self):
        self._call_cmd(["regexp%R1", "rule", "#uname eq node1"])
        self.lib_module.create_plain_with_rule.assert_called_once_with(
            const.RESOURCE_ID_TYPE_REGEXP,
            "R1",
            "#uname eq node1",
            {},
            {},
            set(),
        )
        self.report_processor.set_report_item_preprocessor.assert_called_once()

    def test_resource_id_type_bad(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd("pattern%R1 rule #uname eq node1".split())
        self.assertEqual(
            cm.exception.message,
            "'pattern' is not an allowed type for 'pattern%R1', use regexp, resource",
        )
        self.lib_module.create_plain_with_rule.assert_not_called()
        self.report_processor.set_report_item_preprocessor.assert_not_called()

    @mock.patch("pcs.cli.common.parse_args.deprecation_warning")
    def test_all_options(self, mock_dw):
        self._call_cmd(
            (
                "R1 rule id=id1 constraint-id=id2 score=7 score-attribute=attr "
                "resource-discovery=rd role=r something=anything #uname eq node1"
            ).split(),
            {"force": True},
        )
        self.lib_module.create_plain_with_rule.assert_called_once_with(
            const.RESOURCE_ID_TYPE_PLAIN,
            "R1",
            "something=anything #uname eq node1",
            {"id": "id1", "score": "7", "score-attribute": "attr", "role": "r"},
            {"resource-discovery": "rd", "id": "id2"},
            {reports.codes.FORCE},
        )
        self.report_processor.set_report_item_preprocessor.assert_called_once()
        mock_dw.assert_called_once_with(RULE_ARGV_DEPRECATED)



from unittest import TestCase

from pcs.common.pacemaker.resource.list import (
    CibResourcesDto,
    get_all_resources_ids,
    get_stonith_resources_ids,
)

from pcs_test.tools.resources_dto import (
    ALL_RESOURCES,
    PRIMITIVE_R1,
    PRIMITIVE_R2,
)


class GetAllResourcesIds(TestCase):
    def test_resources(self):
        self.assertEqual(
            get_all_resources_ids(ALL_RESOURCES),
            {
                "R1",
                "R2",
                "R3",
                "R4",
                "R5",
                "R6",
                "R7",
                "S1",
                "S2",
                "G1",
                "G2",
                "B1",
                "B2",
                "G1-clone",
                "R6-clone",
            },
        )

    def test_no_resources(self):
        self.assertEqual(
            get_all_resources_ids(CibResourcesDto([], [], [], [])), set()
        )


class GetStonithResourcesIds(TestCase):
    def test_resources(self):
        self.assertEqual(get_stonith_resources_ids(ALL_RESOURCES), {"S1", "S2"})

    def test_no_resources(self):
        self.assertEqual(
            get_stonith_resources_ids(CibResourcesDto([], [], [], [])), set()
        )

    def test_no_stonith_resources(self):
        self.assertEqual(
            get_stonith_resources_ids(
                CibResourcesDto([PRIMITIVE_R1, PRIMITIVE_R2], [], [], [])
            ),
            set(),
        )



import json
from typing import Mapping, Optional

from pcs.common.file_type_codes import (
    PCS_KNOWN_HOSTS,
    PCS_SETTINGS_CONF,
    FileTypeCode,
)
from pcs.common.types import StringSequence

from pcs_test.tools.command_env.mock_node_communicator import (
    place_multinode_call,
)


class PcsCfgsyncShortcuts:
    _FILETYPE_CODE_TO_LEGACY_NAME_MAP = {
        PCS_KNOWN_HOSTS: "known-hosts",
        PCS_SETTINGS_CONF: "pcs_settings.conf",
    }

    def __init__(self, calls):
        self.__calls = calls

    def set_configs(
        self,
        cluster_name: str = "test99",
        file_contents: Optional[Mapping[FileTypeCode, str]] = None,
        force: bool = False,
        node_labels: Optional[StringSequence] = None,
        communication_list: Optional[Mapping[str, str]] = None,
        name="http.pcs_cfgsync.set_configs",
    ):
        """
        Create a call for sending synced pcs config files to nodes

        cluster_name -- name of the cluster
        file_contents -- contents of files that are sent
        force -- whether the force flag is set in the request
        node_labels -- create success responses from these nodes
        communication_list -- create custom responses
        name -- name of this call
        """
        if file_contents:
            configs = {
                self._FILETYPE_CODE_TO_LEGACY_NAME_MAP[filetype_code]: {
                    "type": "file",
                    "text": content,
                }
                for filetype_code, content in file_contents.items()
            }
        else:
            configs = {}

        output = json.dumps(
            {"status": "ok", "result": dict.fromkeys(configs, "accepted")}
        )

        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/set_configs",
            param_list=[
                (
                    "configs",
                    json.dumps(
                        {
                            "cluster_name": cluster_name,
                            "force": force,
                            "configs": configs,
                        }
                    ),
                )
            ],
            output=output,
        )



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Union

from pcs.common.interface.dto import DataTransferObject


@dataclass(frozen=True)
class CibFencingLevelNodeDto(DataTransferObject):
    id: str
    target: str
    index: int
    devices: list[str]


@dataclass(frozen=True)
class CibFencingLevelRegexDto(DataTransferObject):
    id: str
    target_pattern: str
    index: int
    devices: list[str]


@dataclass(frozen=True)
class CibFencingLevelAttributeDto(DataTransferObject):
    id: str
    target_attribute: str
    target_value: str
    index: int
    devices: list[str]


CibFencingLevel = Union[
    CibFencingLevelNodeDto, CibFencingLevelRegexDto, CibFencingLevelAttributeDto
]


@dataclass(frozen=True)
class CibFencingTopologyDto(DataTransferObject):
    target_node: Sequence[CibFencingLevelNodeDto]
    target_regex: Sequence[CibFencingLevelRegexDto]
    target_attribute: Sequence[CibFencingLevelAttributeDto]



from dataclasses import dataclass
from typing import Optional

from pcs.common.const import (
    PcmkAction,
    PcmkRoleType,
)
from pcs.common.interface.dto import DataTransferObject
from pcs.common.pacemaker.types import (
    CibResourceSetOrdering,
    CibResourceSetOrderType,
)
from pcs.common.types import StringSequence


@dataclass(frozen=True)
class CibResourceSetDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes
    set_id: str
    sequential: Optional[bool]
    require_all: Optional[bool]
    ordering: Optional[CibResourceSetOrdering]
    action: Optional[PcmkAction]
    role: Optional[PcmkRoleType]
    score: Optional[str]
    kind: Optional[CibResourceSetOrderType]
    resources_ids: StringSequence



from dataclasses import dataclass
from typing import (
    Iterable,
    Sequence,
    Union,
)

from pcs.common.interface.dto import DataTransferObject

from .colocation import (
    CibConstraintColocationDto,
    CibConstraintColocationSetDto,
)
from .location import (
    CibConstraintLocationDto,
    CibConstraintLocationSetDto,
)
from .order import (
    CibConstraintOrderDto,
    CibConstraintOrderSetDto,
)
from .ticket import (
    CibConstraintTicketDto,
    CibConstraintTicketSetDto,
)


@dataclass(frozen=True)
class CibConstraintsDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes
    location: Sequence[CibConstraintLocationDto] = tuple()
    location_set: Sequence[CibConstraintLocationSetDto] = tuple()
    colocation: Sequence[CibConstraintColocationDto] = tuple()
    colocation_set: Sequence[CibConstraintColocationSetDto] = tuple()
    order: Sequence[CibConstraintOrderDto] = tuple()
    order_set: Sequence[CibConstraintOrderSetDto] = tuple()
    ticket: Sequence[CibConstraintTicketDto] = tuple()
    ticket_set: Sequence[CibConstraintTicketSetDto] = tuple()


def _get_constraint_ids(
    constraint_dtos: Iterable[
        Union[
            CibConstraintLocationDto,
            CibConstraintLocationSetDto,
            CibConstraintColocationDto,
            CibConstraintColocationSetDto,
            CibConstraintOrderDto,
            CibConstraintOrderSetDto,
            CibConstraintTicketDto,
            CibConstraintTicketSetDto,
        ]
    ],
) -> list[str]:
    return [
        constraint_dto.attributes.constraint_id
        for constraint_dto in constraint_dtos
    ]


def _get_location_rule_ids(
    constraint_dtos: Iterable[
        Union[
            CibConstraintLocationDto,
            CibConstraintLocationSetDto,
        ]
    ],
) -> list[str]:
    return [
        rule_dto.id
        for constraint_dto in constraint_dtos
        for rule_dto in constraint_dto.attributes.rules
    ]


def get_all_constraints_ids(constraints_dto: CibConstraintsDto) -> set[str]:
    return set(
        _get_constraint_ids(constraints_dto.location)
        + _get_constraint_ids(constraints_dto.location_set)
        + _get_constraint_ids(constraints_dto.colocation)
        + _get_constraint_ids(constraints_dto.colocation_set)
        + _get_constraint_ids(constraints_dto.order)
        + _get_constraint_ids(constraints_dto.order_set)
        + _get_constraint_ids(constraints_dto.ticket)
        + _get_constraint_ids(constraints_dto.ticket_set)
    )


def get_all_location_rules_ids(
    constraints_dto: CibConstraintsDto,
) -> set[str]:
    return set(
        _get_location_rule_ids(constraints_dto.location)
        + _get_location_rule_ids(constraints_dto.location_set)
    )


def get_all_location_constraints_ids(
    constraints_dto: CibConstraintsDto,
) -> set[str]:
    return set(
        _get_constraint_ids(constraints_dto.location)
        + _get_constraint_ids(constraints_dto.location_set)
    )



from pcs.daemon import ruby_pcsd
from pcs.daemon.app.auth import (
    NotAuthorizedException,
    UnixSocketAuthProvider,
)
from pcs.daemon.app.common import (
    LegacyApiHandler,
    RoutesType,
)
from pcs.daemon.app.sinatra_common import SinatraMixin
from pcs.daemon.app.ui_common import AjaxMixin
from pcs.lib.auth.provider import AuthProvider
from pcs.lib.auth.types import AuthUser


class SinatraAjaxProtected(LegacyApiHandler, SinatraMixin, AjaxMixin):
    """
    SinatraAjaxProtected handles urls that calls the ajax Sinatra GUI functions.
    It allows to use this urls only for ajax calls.
    """

    _auth_provider: UnixSocketAuthProvider

    def initialize(
        self,
        ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
        auth_provider: AuthProvider,
    ) -> None:
        self.initialize_sinatra(ruby_pcsd_wrapper)
        self._auth_provider = UnixSocketAuthProvider(self, auth_provider)

    def prepare(self) -> None:
        if not self.is_ajax:
            raise self.unauthorized()

    async def _handle_request(self):
        try:
            auth_user = await self._get_auth_user()
        except NotAuthorizedException as e:
            raise self.unauthorized() from e

        # Webui version needs to deal with extra session operation here.
        self._after_get_auth_user_hook(auth_user)

        result = await self.ruby_pcsd_wrapper.request(
            auth_user,
            self.request,
        )
        self.send_sinatra_result(result)

    async def _get_auth_user(self):
        return await self._auth_provider.auth_by_socket_user()

    def _after_get_auth_user_hook(self, auth_user: AuthUser):
        pass


def get_routes(
    auth_provider: AuthProvider,
    ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
) -> RoutesType:
    return [
        # The protection by session was moved from ruby code to python code
        # (tornado).
        (
            r"/.*",
            SinatraAjaxProtected,
            dict(
                ruby_pcsd_wrapper=ruby_pcsd_wrapper,
                auth_provider=auth_provider,
            ),
        ),
    ]



from dataclasses import dataclass

from pcs.common.file_type_codes import FileTypeCode
from pcs.common.interface.dto import DataTransferObject


@dataclass(frozen=True)
class SyncConfigsDto(DataTransferObject):
    cluster_name: str
    configs: dict[FileTypeCode, str]



from unittest import (
    TestCase,
    mock,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.constraint import command as constraint_command

from pcs_test.tools.constraints_dto import get_all_constraints
from pcs_test.tools.custom_mock import RuleInEffectEvalMock
from pcs_test.tools.misc import dict_to_modifiers


class TestRemoveConstraint(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["constraint", "cib"])
        self.cib = mock.Mock(spec_set=["remove_elements"])
        self.constraint = mock.Mock(spec_set=["get_config"])
        self.constraint.get_config.return_value = get_all_constraints(
            RuleInEffectEvalMock({})
        )
        self.lib.cib = self.cib
        self.lib.constraint = self.constraint

    def _call_cmd(self, argv, modifiers=None):
        constraint_command.remove(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.constraint.get_config.assert_not_called()
        self.cib.remove_elements.assert_not_called()

    def test_duplicate_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["id3", "id1", "id2", "id1", "id2"])
        self.assertEqual(
            cm.exception.message, "duplicate arguments: 'id1', 'id2'"
        )
        self.constraint.get_config.assert_not_called()
        self.cib.remove_elements.assert_not_called()

    def test_unsupported_modifier(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["id1", "id2", "id1", "id2"], {"force": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--force' is not supported in this command",
        )
        self.constraint.get_config.assert_not_called()
        self.cib.remove_elements.assert_not_called()

    def test_not_constraint_or_rule_ids(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["id1", "id2", "location-R7-localhost-INFINITY"])
        self.assertEqual(
            cm.exception.message,
            "Unable to find constraints or constraint rules: 'id1', 'id2'",
        )
        self.constraint.get_config.assert_called_once_with(evaluate_rules=False)
        self.cib.remove_elements.assert_not_called()

    def test_constraint_or_rule_ids(self):
        constraint_or_rule_ids = [
            "location-R7-localhost-INFINITY",
            "loc_constr_with_expired_rule-rule",
            "loc_constr_with_not_expired_rule-1-rule",
            "colocation-G1-clone-R6-clone--100",
            "order_set_B2R6-cloneSe",
            "ticket_set_R7B2G2",
        ]
        self._call_cmd(constraint_or_rule_ids)
        self.constraint.get_config.assert_called_once_with(evaluate_rules=False)
        self.cib.remove_elements.assert_called_once_with(constraint_or_rule_ids)



from functools import partial

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.pacemaker.constraint import CibConstraintsDto
from pcs.lib import validate
from pcs.lib.cib.constraint import (
    colocation,
    common,
    constraint,
    location,
    order,
    resource_set,
    ticket,
)
from pcs.lib.cib.rule.in_effect import get_rule_evaluator
from pcs.lib.cib.tools import (
    ElementNotFound,
    get_constraints,
    get_element_by_id,
)
from pcs.lib.env import LibraryEnvironment


# This is an extracted part of lib commands for creating set constraints for
# the purposes of code deduplication. It is not meant to be part of API.
# DEPRECATED
def create_with_set(
    tag_name,
    prepare_options,
    env,
    resource_set_list,
    constraint_options,
    resource_in_clone_alowed=False,
    duplication_alowed=False,
    duplicate_check=None,
):
    """
    string tag_name is constraint tag name
    callable prepare_options takes
        cib(Element), options(dict), resource_set_list and return corrected
        options or if options not usable raises error
    env is library environment
    list resource_set_list is description of resource set, for example:
        {"ids": ["A", "B"], "options": {"sequential": "true"}},
    dict constraint_options is base for building attributes of constraint tag
    bool resource_in_clone_alowed flag for allowing to reference id which is
        in tag clone or master
    bool duplication_alowed flag for allowing create duplicate element
    callable duplicate_check takes two elements and decide if they are
        duplicates
    """
    cib = env.get_cib()

    find_valid_resource_id = partial(
        constraint.find_valid_resource_id,
        env.report_processor,
        cib,
        resource_in_clone_alowed,
    )

    constraint_section = get_constraints(cib)
    constraint_element = constraint.create_with_set(
        constraint_section,
        tag_name,
        options=prepare_options(cib, constraint_options, resource_set_list),
        resource_set_list=[
            resource_set.prepare_set(
                find_valid_resource_id, resource_set_item, env.report_processor
            )
            for resource_set_item in resource_set_list
        ],
    )

    if not duplicate_check:
        duplicate_check = constraint.have_duplicate_resource_sets

    constraint.check_is_without_duplication(
        env.report_processor,
        constraint_section,
        constraint_element,
        are_duplicate=duplicate_check,
        duplication_allowed=duplication_alowed,
    )

    env.push_cib()


def get_config(
    env: LibraryEnvironment,
    evaluate_rules: bool = False,
) -> CibConstraintsDto:
    cib = env.get_cib()
    constraints_el = get_constraints(cib)
    rule_evaluator = get_rule_evaluator(
        cib, env.cmd_runner(), env.report_processor, evaluate_rules
    )
    location_constraints, location_set_constraints = location.get_all_as_dtos(
        constraints_el, rule_evaluator
    )
    (
        colocation_constraints,
        colocation_set_constraints,
    ) = colocation.get_all_as_dtos(constraints_el, rule_evaluator)
    order_constraints, order_set_constraints = order.get_all_as_dtos(
        constraints_el
    )
    ticket_constraints, ticket_set_constraints = ticket.get_all_as_dtos(
        constraints_el
    )
    return CibConstraintsDto(
        location=location_constraints,
        location_set=location_set_constraints,
        colocation=colocation_constraints,
        colocation_set=colocation_set_constraints,
        order=order_constraints,
        order_set=order_set_constraints,
        ticket=ticket_constraints,
        ticket_set=ticket_set_constraints,
    )


# This is an extracted part of lib commands for creating set constraints for
# the purposes of code deduplication. It is not meant to be part of API.
def _load_resource_set_list(
    cib: _Element,
    report_processor: reports.ReportProcessor,
    rsc_set_list: common.CmdInputResourceSetList,
    option_value_normalization: validate.TypeNormalizeFunc,
) -> common.CmdInputResourceSetLoadedList:
    """
    Prepare resource sets definition for further processing in commands

    cib -- loaded CIB
    rsc_set_list -- resources set definition passed as a command input
    option_value_normalization -- function for normalizing set options values
    """
    resource_set_el_list: common.CmdInputResourceSetLoadedList = []
    for input_set_def in rsc_set_list:
        el_list = []
        for resource_id in input_set_def["ids"]:
            try:
                el_list.append(get_element_by_id(cib, resource_id))
            except ElementNotFound:
                report_processor.report(
                    reports.ReportItem.error(
                        reports.messages.IdNotFound(resource_id, [])
                    )
                )
        resource_set_el_list.append(
            dict(
                constrained_elements=el_list,
                options=validate.values_to_pairs(
                    input_set_def["options"], option_value_normalization
                ),
            )
        )
    return resource_set_el_list


# This is an extracted part of lib commands for creating set constraints for
# the purposes of code deduplication. It is not meant to be part of API.
def _primitive_resource_set_list(
    rsc_set_list: common.CmdInputResourceSetLoadedList,
) -> common.CmdInputResourceSetList:
    """
    Inverse transformation to _load_resource_set_list
    """
    return [
        dict(
            ids=[
                str(el.attrib["id"]) for el in rsc_set["constrained_elements"]
            ],
            options=validate.pairs_to_values(rsc_set["options"]),
        )
        for rsc_set in rsc_set_list
    ]



from unittest import (
    TestCase,
    mock,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import InputModifiers
from pcs.cli.resource import command
from pcs.common import reports
from pcs.lib.errors import LibraryError

from pcs_test.tools.misc import dict_to_modifiers
from pcs_test.tools.resources_dto import ALL_RESOURCES

CANNOT_STOP_RESOURCES_ERROR_REPORT = reports.ReportItem.error(
    reports.messages.CannotRemoveResourcesNotStopped(["R1"])
)
NO_STONITH_WOULD_BE_LEFT_ERROR_REPORT = reports.ReportItem.error(
    reports.messages.NoStonithMeansWouldBeLeft()
)
FORCEABLE_ERROR_REPORT = reports.ReportItem.error(
    reports.messages.NoStonithMeansWouldBeLeft(), force_code=reports.codes.FORCE
)
INFO_REPORT = reports.ReportItem.info(
    reports.messages.CibRemoveDependantElements({})
)


class RemoveResourceBase:
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cib", "cluster", "env", "resource"])

        self.lib.env = mock.Mock(spec_set=["report_processor"])
        self.lib.cib = mock.Mock(spec_set=["remove_elements"])
        self.lib.cluster = mock.Mock(spec_set=["wait_for_pcmk_idle"])
        self.lib.resource = mock.Mock(
            spec_set=["stop", "get_configured_resources"]
        )
        self.lib.resource.get_configured_resources.return_value = ALL_RESOURCES

    def _call_cmd(self, argv, modifiers=None):
        command.remove(self.lib, argv, dict_to_modifiers(modifiers or {}))

    def assert_lib_calls(self, expected_calls):
        self.lib.assert_has_calls(expected_calls)
        self.assertEqual(len(self.lib.mock_calls), len(expected_calls))

    def test_no_args(self, mock_process_lib_reports, mock_reports):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.assert_lib_calls([])
        mock_reports.assert_not_called()
        mock_process_lib_reports.assert_not_called()

    def test_duplicate_args(self, mock_process_lib_reports, mock_reports):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["R1", "R1", "R2", "R3", "R2"])
        self.assertEqual(
            cm.exception.message, "duplicate arguments: 'R1', 'R2'"
        )
        self.assert_lib_calls([])
        mock_reports.assert_not_called()
        mock_process_lib_reports.assert_not_called()

    def test_not_resource_id(self, mock_process_lib_reports, mock_reports):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["nonexistent"])

        self.assertEqual(
            cm.exception.message, "Unable to find resource: 'nonexistent'"
        )
        self.assert_lib_calls([mock.call.resource.get_configured_resources()])
        mock_reports.assert_not_called()
        mock_process_lib_reports.assert_not_called()

    def test_stonith_id(self, mock_process_lib_reports, mock_reports):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["S1"])

        self.assertEqual(
            cm.exception.message,
            (
                "This command cannot remove stonith resource: 'S1'. "
                "Use 'pcs stonith remove' instead."
            ),
        )
        self.assert_lib_calls([mock.call.resource.get_configured_resources()])
        mock_reports.assert_not_called()
        mock_process_lib_reports.assert_not_called()

    def test_multiple_stonith_ids(self, mock_process_lib_reports, mock_reports):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["S1", "R1", "R2", "R3", "S2"])

        self.assertEqual(
            cm.exception.message,
            (
                "This command cannot remove stonith resources: 'S1', 'S2'. "
                "Use 'pcs stonith remove' instead."
            ),
        )
        self.assert_lib_calls([mock.call.resource.get_configured_resources()])
        mock_reports.assert_not_called()
        mock_process_lib_reports.assert_not_called()

    def test_remove_one(self, mock_process_library_reports, mock_reports):
        mock_reports.return_value = [INFO_REPORT]
        self._call_cmd(["R1"])

        self.assert_lib_calls(
            [
                mock.call.resource.get_configured_resources(),
                mock.call.cib.remove_elements({"R1"}),
            ]
        )
        mock_process_library_reports.assert_called_once_with(
            [INFO_REPORT], include_debug=False
        )

    def test_remove_one_with_debug(
        self, mock_process_library_reports, mock_reports
    ):
        mock_reports.return_value = [INFO_REPORT]
        self._call_cmd(["R1"], {"debug": True})

        self.assert_lib_calls(
            [
                mock.call.resource.get_configured_resources(),
                mock.call.cib.remove_elements({"R1"}),
            ]
        )
        mock_process_library_reports.assert_called_once_with(
            [INFO_REPORT], include_debug=True
        )

    def test_remove_multiple(self, mock_process_library_reports, mock_reports):
        mock_reports.return_value = []
        self._call_cmd(["R1", "R2", "R3"])

        self.assert_lib_calls(
            [
                mock.call.resource.get_configured_resources(),
                mock.call.cib.remove_elements({"R1", "R2", "R3"}),
            ]
        )
        mock_process_library_reports.assert_not_called()

    def test_dont_stop_me_now(self, mock_process_library_reports, mock_reports):
        self._call_cmd(["R1"], {"no-stop": True})

        self.assert_lib_calls(
            [
                mock.call.resource.get_configured_resources(),
                mock.call.cib.remove_elements({"R1"}, set()),
            ]
        )
        mock_reports.assert_not_called()
        mock_process_library_reports.assert_not_called()

    def test_force_dont_stop_me_now(
        self, mock_process_library_reports, mock_reports
    ):
        self._call_cmd(["R1"], {"force": True, "no-stop": True})

        self.assert_lib_calls(
            [
                mock.call.resource.get_configured_resources(),
                mock.call.cib.remove_elements({"R1"}, {reports.codes.FORCE}),
            ]
        )
        mock_reports.assert_not_called()
        mock_process_library_reports.assert_not_called()

    def test_remove_not_stopped(
        self, mock_process_library_reports, mock_reports
    ):
        self.lib.cib.remove_elements.side_effect = [LibraryError(), None]
        mock_reports.return_value = [CANNOT_STOP_RESOURCES_ERROR_REPORT]

        self._call_cmd(["R1"])

        resource_ids = {"R1"}
        force_flags = set()
        self.assert_lib_calls(
            [
                mock.call.resource.get_configured_resources(),
                mock.call.cib.remove_elements(resource_ids),
                mock.call.resource.stop(resource_ids, force_flags),
                mock.call.cluster.wait_for_pcmk_idle(None),
                mock.call.cib.remove_elements(resource_ids, force_flags),
            ]
        )
        mock_process_library_reports.assert_not_called()

    def test_remove_more_errors(
        self, mock_process_library_reports, mock_reports
    ):
        self.lib.cib.remove_elements.side_effect = [LibraryError(), None]
        mock_reports.return_value = [
            NO_STONITH_WOULD_BE_LEFT_ERROR_REPORT,
            CANNOT_STOP_RESOURCES_ERROR_REPORT,
        ]

        self.assertRaises(LibraryError, lambda: self._call_cmd(["R1"]))

        self.assert_lib_calls(
            [
                mock.call.resource.get_configured_resources(),
                mock.call.cib.remove_elements({"R1"}),
            ]
        )
        mock_process_library_reports.assert_called_once_with(
            [NO_STONITH_WOULD_BE_LEFT_ERROR_REPORT],
            include_debug=False,
            exit_on_error=False,
        )

    def test_remove_more_errors_debug(
        self, mock_process_library_reports, mock_reports
    ):
        self.lib.cib.remove_elements.side_effect = [LibraryError(), None]
        mock_reports.return_value = [
            NO_STONITH_WOULD_BE_LEFT_ERROR_REPORT,
            CANNOT_STOP_RESOURCES_ERROR_REPORT,
        ]

        self.assertRaises(
            LibraryError, lambda: self._call_cmd(["R1"], {"debug": True})
        )

        self.assert_lib_calls(
            [
                mock.call.resource.get_configured_resources(),
                mock.call.cib.remove_elements({"R1"}),
            ]
        )
        mock_process_library_reports.assert_called_once_with(
            [NO_STONITH_WOULD_BE_LEFT_ERROR_REPORT],
            include_debug=True,
            exit_on_error=False,
        )

    def test_mutually_exclusive_options(
        self, mock_process_library_reports, mock_reports
    ):
        with self.assertRaises(CmdLineInputError) as cm:
            command.remove(
                self.lib,
                ["R1"],
                InputModifiers({"-f": "foo", "--no-stop": True}),
            )

        self.assert_lib_calls([])
        self.assertEqual(
            cm.exception.message, "Only one of '--no-stop', '-f' can be used"
        )
        mock_reports.assert_not_called()
        mock_process_library_reports.assert_not_called()


@mock.patch(
    "pcs.common.reports.processor.ReportProcessorInMemory.reports",
    new_callable=mock.PropertyMock,
)
@mock.patch("pcs.cli.resource.command.process_library_reports")
class RemoveResource(RemoveResourceBase, TestCase):
    @mock.patch("pcs.cli.resource.command.deprecation_warning")
    def test_remove_force(
        self, mock_deprecation_warning, mock_process_lib_reports, mock_reports
    ):
        self._call_cmd(["R1"], {"force": True})

        mock_deprecation_warning.assert_called_once()
        self.assert_lib_calls(
            [
                mock.call.resource.get_configured_resources(),
                mock.call.cib.remove_elements({"R1"}, {reports.codes.FORCE}),
            ]
        )
        mock_reports.assert_not_called()
        mock_process_lib_reports.assert_not_called()


@mock.patch(
    "pcs.common.reports.processor.ReportProcessorInMemory.reports",
    new_callable=mock.PropertyMock,
)
@mock.patch("pcs.cli.resource.command.process_library_reports")
class RemoveResourceFuture(RemoveResourceBase, TestCase):
    def _call_cmd(self, argv, modifiers=None):
        default_modifiers = {"future": True}
        command.remove(
            self.lib,
            argv,
            dict_to_modifiers(
                modifiers | default_modifiers
                if modifiers
                else default_modifiers
            ),
        )

    def test_remove_force_more_errors_forceable(
        self, mock_process_lib_reports, mock_reports
    ):
        self.lib.cib.remove_elements.side_effect = [LibraryError(), None]
        mock_reports.return_value = [
            FORCEABLE_ERROR_REPORT,
            CANNOT_STOP_RESOURCES_ERROR_REPORT,
        ]

        self._call_cmd(["R1"], {"force": True})

        resource_ids = {"R1"}
        force_codes = {reports.codes.FORCE}
        self.assert_lib_calls(
            [
                mock.call.resource.get_configured_resources(),
                mock.call.cib.remove_elements(resource_ids),
                mock.call.resource.stop(resource_ids, force_codes),
                mock.call.cluster.wait_for_pcmk_idle(None),
                mock.call.cib.remove_elements(resource_ids, force_codes),
            ]
        )
        mock_process_lib_reports.assert_not_called()

    def test_remove_force_more_errors_not_forceable(
        self, mock_process_lib_reports, mock_reports
    ):
        self.lib.cib.remove_elements.side_effect = [LibraryError(), None]
        mock_reports.return_value = [
            FORCEABLE_ERROR_REPORT,
            NO_STONITH_WOULD_BE_LEFT_ERROR_REPORT,
            CANNOT_STOP_RESOURCES_ERROR_REPORT,
        ]

        self.assertRaises(
            LibraryError,
            lambda: self._call_cmd(["R1"], {"force": True}),
        )
        self.assert_lib_calls(
            [
                mock.call.resource.get_configured_resources(),
                mock.call.cib.remove_elements({"R1"}),
            ]
        )
        mock_process_lib_reports.assert_called_once_with(
            [
                reports.item.ReportItem.warning(
                    reports.messages.NoStonithMeansWouldBeLeft()
                ),
                NO_STONITH_WOULD_BE_LEFT_ERROR_REPORT,
            ],
            include_debug=False,
            exit_on_error=False,
        )



from pcs.common import reports

from pcs_test.tools import fixture


class CrmAttributeLoadMetadataMixin:
    def load_fake_agent_metadata(self):
        self.config.runner.pcmk.is_crm_attribute_list_options_supported(
            is_supported=True
        )
        self.config.runner.pcmk.load_crm_attribute_metadata()


class CrmAttributeMetadataErrorMixin:
    _load_cib_when_metadata_error = None

    def metadata_error_command(self):
        raise NotImplementedError

    def _metadata_error(
        self,
        agent="cluster-options",
        stdout=None,
        stderr="",
        reason=None,
        returncode=2,
        unsupported_version=False,
    ):
        if self._load_cib_when_metadata_error:
            self.config.runner.cib.load()
        self.config.runner.pcmk.is_crm_attribute_list_options_supported(
            is_supported=True
        )
        self.config.runner.pcmk.load_crm_attribute_metadata(
            agent_name=agent,
            stdout=stdout,
            stderr=stderr,
            returncode=returncode,
        )
        self.env_assist.assert_raise_library_error(self.metadata_error_command)
        if unsupported_version:
            report = fixture.error(
                reports.codes.AGENT_IMPLEMENTS_UNSUPPORTED_OCF_VERSION,
                agent=f"__pcmk_internal:{agent}",
                ocf_version="5.2",
                supported_versions=["1.0", "1.1"],
            )
        else:
            report = fixture.error(
                reports.codes.UNABLE_TO_GET_AGENT_METADATA,
                agent=agent,
                reason="error" if reason is None else reason,
            )
        self.env_assist.assert_reports([report])

    def test_metadata_error_returncode(self):
        stdout = """
            <pacemaker-result api-version="2.38" request="crm_attribute">
                <status code="2" message="ERROR" />
            </pacemaker-result>
        """
        self._metadata_error(stdout=stdout, reason="ERROR")

    def test_metadata_error_xml_syntax_error(self):
        stdout = "not an xml"
        stderr = "syntax error"
        self._metadata_error(
            stdout=stdout, stderr=stderr, reason=f"{stderr}\n{stdout}"
        )

    def test_metadata_error_invalid_schema(self):
        stdout = "<xml/>"
        stderr = "invalid schema"
        self._metadata_error(
            stdout=stdout, stderr=stderr, reason=f"{stderr}\n{stdout}"
        )

    def test_metadata_error_invalid_version(self):
        stdout = """
            <pacemaker-result api-version="2.38" request="crm_attribute">
                <resource-agent name="cluster-options">
                    <version>5.2</version>
                    <parameters>
                        <parameter name="parameter-name" advanced="0"
                                generated="0">
                            <longdesc lang="en">longdesc</longdesc>
                            <shortdesc lang="en">shortdesc</shortdesc>
                            <content type="string"/>
                        </parameter>
                    </parameters>
                </resource-agent>
                <status code="0" message="OK" />
            </pacemaker-result>
        """
        self._metadata_error(
            stdout=stdout, returncode=0, unsupported_version=True
        )

    def test_facade_crm_attribute_metadata_not_supported(self):
        if self._load_cib_when_metadata_error:
            self.config.runner.cib.load()
        self.config.runner.pcmk.is_crm_attribute_list_options_supported(
            is_supported=False
        )
        self.env_assist.assert_raise_library_error(self.metadata_error_command)
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.CLUSTER_OPTIONS_METADATA_NOT_SUPPORTED
                )
            ]
        )



from unittest import TestCase

from lxml import etree

from pcs.common import const, reports
from pcs.lib.cib.constraint import ticket
from pcs.lib.cib.tools import IdProvider, Version

from pcs_test.tier0.lib.cib.constraint.test_common import (
    DuplicatesCheckerTestBase,
)
from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)
from pcs_test.tools.xml import etree_to_str, str_to_etree


class IsTicketConstraint(TestCase):
    def test_ticket_constraint_true(self):
        self.assertTrue(
            ticket.is_ticket_constraint(etree.Element("rsc_ticket"))
        )

    def test_ticket_constraint_false(self):
        self.assertFalse(ticket.is_ticket_constraint(etree.Element("ticket")))


class ValidateCreatePlain(TestCase):
    def setUp(self):
        cib = str_to_etree("""
            <resources>
                <primitive id="R">
                    <meta_attributes id="R-meta" />
                </primitive>
                <clone id="C">
                    <primitive id="CR" />
                </clone>
            </resources>
        """)
        self.el_primitive = cib.xpath(".//*[@id='R']")[0]
        self.el_meta = cib.xpath(".//*[@id='R-meta']")[0]
        self.el_clone = cib.xpath(".//*[@id='C']")[0]
        self.el_cloned_primitive = cib.xpath(".//*[@id='CR']")[0]
        self.id_provider = IdProvider(cib)

    def test_resource_not_a_resource(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider, "T", self.el_meta, {}, False
            ),
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R-meta",
                    expected_types=[
                        "bundle",
                        "clone",
                        "group",
                        "master",
                        "primitive",
                    ],
                    current_type="meta_attributes",
                ),
            ],
        )

    def test_resource_in_a_clone(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider, "T", self.el_cloned_primitive, {}, False
            ),
            [
                fixture.error(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    force_code=reports.codes.FORCE,
                    resource_id="CR",
                    parent_type="clone",
                    parent_id="C",
                ),
            ],
        )

    def test_resource_in_a_clone_forced(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider, "T", self.el_cloned_primitive, {}, True
            ),
            [
                fixture.warn(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    resource_id="CR",
                    parent_type="clone",
                    parent_id="C",
                ),
            ],
        )

    def test_clone(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider, "T", self.el_clone, {}, False
            ),
            [],
        )

    def test_ticket_missing(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider, "", self.el_primitive, {}, False
            ),
            [
                fixture.error(
                    reports.codes.REQUIRED_OPTIONS_ARE_MISSING,
                    option_names=["ticket"],
                    option_type=None,
                )
            ],
        )

    def test_ticket_not_valid(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider, "my_ticket", self.el_primitive, {}, False
            ),
            [
                fixture.error(
                    reports.codes.BOOTH_TICKET_NAME_INVALID,
                    ticket_name="my_ticket",
                )
            ],
        )

    def test_id_not_valid(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider, "T", self.el_primitive, {"id": "123"}, False
            ),
            [
                fixture.error(
                    reports.codes.INVALID_ID_BAD_CHAR,
                    id="123",
                    id_description="constraint id",
                    invalid_character="1",
                    is_first_char=True,
                )
            ],
        )

    def test_id_already_used(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider, "T", self.el_primitive, {"id": "R"}, False
            ),
            [
                fixture.error(
                    reports.codes.ID_ALREADY_EXISTS,
                    id="R",
                )
            ],
        )

    def test_role_not_valid(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider,
                "T",
                self.el_primitive,
                {"rsc-role": "bad"},
                False,
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="role",
                    option_value="bad",
                    allowed_values=(
                        const.PCMK_ROLE_STOPPED,
                        const.PCMK_ROLE_STARTED,
                        const.PCMK_ROLE_PROMOTED,
                        const.PCMK_ROLE_UNPROMOTED,
                    ),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_role_legacy(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider,
                "T",
                self.el_primitive,
                {"rsc-role": const.PCMK_ROLE_PROMOTED_LEGACY},
                False,
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="role",
                    option_value=const.PCMK_ROLE_PROMOTED_LEGACY,
                    allowed_values=(
                        const.PCMK_ROLE_STOPPED,
                        const.PCMK_ROLE_STARTED,
                        const.PCMK_ROLE_PROMOTED,
                        const.PCMK_ROLE_UNPROMOTED,
                    ),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_loss_policy_not_valid(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider,
                "T",
                self.el_primitive,
                {"loss-policy": "bad"},
                False,
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="loss-policy",
                    option_value="bad",
                    allowed_values=("fence", "stop", "freeze", "demote"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_option_names(self):
        assert_report_item_list_equal(
            ticket.validate_create_plain(
                self.id_provider,
                "T",
                self.el_primitive,
                {
                    "id": "const",
                    "loss-policy": "stop",
                    "rsc-role": const.PCMK_ROLE_PROMOTED,
                    "unknown": "option",
                },
                False,
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["unknown"],
                    allowed=["id", "loss-policy", "rsc-role"],
                    option_type=None,
                    allowed_patterns=[],
                ),
            ],
        )


class CreatePlain(TestCase):
    def setUp(self):
        cib = str_to_etree(
            """
            <cib>
                <resources>
                    <primitive id="resource1" />
                </resources>
                <constraints />
            </cib>
            """
        )
        self.el_constraints = cib.find("constraints")
        self.id_provider = IdProvider(cib)

    def test_minimal_options(self):
        ticket.create_plain(
            self.el_constraints,
            self.id_provider,
            Version(3, 8, 0),
            "ticket1",
            "resource1",
            {},
        )
        assert_xml_equal(
            """
            <constraints>
                <rsc_ticket id="ticket-ticket1-resource1" ticket="ticket1"
                    rsc="resource1"
                />
            </constraints>
            """,
            etree_to_str(self.el_constraints),
        )

    def test_all_options(self):
        ticket.create_plain(
            self.el_constraints,
            self.id_provider,
            Version(3, 8, 0),
            "ticket1",
            "resource1",
            {
                "id": "my-id",
                "loss-policy": "freeze",
                "rsc-role": const.PCMK_ROLE_UNPROMOTED,
            },
        )
        assert_xml_equal(
            """
            <constraints>
                <rsc_ticket id="my-id" loss-policy="freeze" rsc="resource1"
                    rsc-role="Unpromoted" ticket="ticket1"
                />
            </constraints>
            """,
            etree_to_str(self.el_constraints),
        )

    def test_build_id_with_role(self):
        ticket.create_plain(
            self.el_constraints,
            self.id_provider,
            Version(3, 8, 0),
            "ticket1",
            "resource1",
            {"rsc-role": const.PCMK_ROLE_PROMOTED},
        )
        assert_xml_equal(
            """
            <constraints>
                <rsc_ticket id="ticket-ticket1-resource1-{role}"
                    ticket="ticket1" rsc="resource1" rsc-role="{role}"
                />
            </constraints>
            """.format(role=const.PCMK_ROLE_PROMOTED),
            etree_to_str(self.el_constraints),
        )

    def test_build_id_with_role_legacy(self):
        ticket.create_plain(
            self.el_constraints,
            self.id_provider,
            Version(3, 6, 0),
            "ticket1",
            "resource1",
            {"rsc-role": const.PCMK_ROLE_PROMOTED},
        )
        assert_xml_equal(
            """
            <constraints>
                <rsc_ticket id="ticket-ticket1-resource1-{role}"
                    ticket="ticket1" rsc="resource1" rsc-role="{role}"
                />
            </constraints>
            """.format(role=const.PCMK_ROLE_PROMOTED_LEGACY),
            etree_to_str(self.el_constraints),
        )


class ValidateCreateWithSet(TestCase):
    def setUp(self):
        cib = str_to_etree("""
            <resources>
                <primitive id="R1">
                    <meta_attributes id="R1-meta" />
                </primitive>
                <primitive id="R2">
                    <meta_attributes id="R2-meta" />
                </primitive>
                <primitive id="R3" />
                <primitive id="R4" />
                <clone id="C">
                    <primitive id="CR" />
                </clone>
            </resources>
        """)
        self.el_primitives = cib.xpath("./primitive")
        self.el_metas = cib.xpath(".//meta_attributes")
        self.el_clone = cib.xpath(".//*[@id='C']")[0]
        self.el_cloned_primitive = cib.xpath(".//*[@id='CR']")[0]
        self.id_provider = IdProvider(cib)

    @staticmethod
    def set(*args, **kwargs):
        if "require_all" in kwargs:
            kwargs["require-all"] = kwargs["require_all"]
            del kwargs["require_all"]
        return dict(constrained_elements=args, options=kwargs)

    def test_empty_rsc_set_list(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider, [], {"ticket": "T"}, False
            ),
            [
                fixture.error(reports.codes.EMPTY_RESOURCE_SET_LIST),
            ],
        )

    def test_empty_rsc_set(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [self.set(*self.el_primitives[:2]), self.set()],
                {"ticket": "T"},
                False,
            ),
            [
                fixture.error(reports.codes.EMPTY_RESOURCE_SET),
            ],
        )

    def test_resource_not_a_resource(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [
                    self.set(self.el_primitives[0], self.el_metas[0]),
                    self.set(self.el_metas[1], self.el_primitives[1]),
                ],
                {"ticket": "T"},
                False,
            ),
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R1-meta",
                    expected_types=[
                        "bundle",
                        "clone",
                        "group",
                        "master",
                        "primitive",
                    ],
                    current_type="meta_attributes",
                ),
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R2-meta",
                    expected_types=[
                        "bundle",
                        "clone",
                        "group",
                        "master",
                        "primitive",
                    ],
                    current_type="meta_attributes",
                ),
            ],
        )

    def test_resource_in_a_clone(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [self.set(self.el_cloned_primitive)],
                {"ticket": "T"},
                False,
            ),
            [
                fixture.error(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    force_code=reports.codes.FORCE,
                    resource_id="CR",
                    parent_type="clone",
                    parent_id="C",
                ),
            ],
        )

    def test_resource_in_a_clone_forced(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [self.set(self.el_cloned_primitive)],
                {"ticket": "T"},
                True,
            ),
            [
                fixture.warn(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    resource_id="CR",
                    parent_type="clone",
                    parent_id="C",
                ),
            ],
        )

    def test_clone(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [self.set(self.el_clone)],
                {"ticket": "T"},
                False,
            ),
            [],
        )

    def test_ticket_missing(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [self.set(self.el_primitives[0])],
                {},
                False,
            ),
            [
                fixture.error(
                    reports.codes.REQUIRED_OPTIONS_ARE_MISSING,
                    option_names=["ticket"],
                    option_type=None,
                )
            ],
        )

    def test_ticket_not_valid(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [self.set(self.el_primitives[0])],
                {"ticket": "my_ticket"},
                False,
            ),
            [
                fixture.error(
                    reports.codes.BOOTH_TICKET_NAME_INVALID,
                    ticket_name="my_ticket",
                )
            ],
        )

    def test_id_not_valid(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [self.set(self.el_primitives[0])],
                {"id": "123", "ticket": "ticket"},
                False,
            ),
            [
                fixture.error(
                    reports.codes.INVALID_ID_BAD_CHAR,
                    id="123",
                    id_description="constraint id",
                    invalid_character="1",
                    is_first_char=True,
                )
            ],
        )

    def test_id_already_used(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [self.set(self.el_primitives[0])],
                {"id": "R1", "ticket": "ticket"},
                False,
            ),
            [
                fixture.error(
                    reports.codes.ID_ALREADY_EXISTS,
                    id="R1",
                )
            ],
        )

    def test_loss_policy_not_valid(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [self.set(self.el_primitives[0])],
                {"loss-policy": "bad", "ticket": "ticket"},
                False,
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="loss-policy",
                    option_value="bad",
                    allowed_values=("fence", "stop", "freeze", "demote"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )

    def test_option_names(self):
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [self.set(self.el_primitives[0])],
                {"ticket": "ticket", "bad_option": "value"},
                False,
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["bad_option"],
                    allowed=["id", "loss-policy", "ticket"],
                    option_type=None,
                    allowed_patterns=[],
                )
            ],
        )

    def test_set_options(self):
        pcmk_bool = (
            "a pacemaker boolean value: '0', '1', 'false', 'n', 'no', 'off', "
            "'on', 'true', 'y', 'yes'"
        )
        assert_report_item_list_equal(
            ticket.validate_create_with_set(
                self.id_provider,
                [
                    self.set(
                        self.el_primitives[0],
                        action="bad-action",
                        require_all="bad-require",
                        role=const.PCMK_ROLE_PROMOTED_LEGACY,
                        sequential="bad-sequential",
                        bad_option="value",
                    )
                ],
                {"ticket": "ticket"},
                False,
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["bad_option"],
                    allowed=["action", "require-all", "role", "sequential"],
                    option_type="set",
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="action",
                    option_value="bad-action",
                    allowed_values=("start", "stop", "promote", "demote"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="require-all",
                    option_value="bad-require",
                    allowed_values=pcmk_bool,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="role",
                    option_value=const.PCMK_ROLE_PROMOTED_LEGACY,
                    allowed_values=(
                        "Stopped",
                        "Started",
                        "Promoted",
                        "Unpromoted",
                    ),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="sequential",
                    option_value="bad-sequential",
                    allowed_values=pcmk_bool,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class CreateWithSet:
    # pcs_test.tier0.lib.cib.constraint.test_common.CreateConstraintWithSetTest
    pass


class DuplicatesCheckerTicketPlain(DuplicatesCheckerTestBase):
    cib = etree.fromstring(
        """
        <constraints>
            <rsc_ticket id="C1" ticket="T1" rsc="R1" />
            <rsc_ticket id="C2" ticket="T1" rsc="R1" />
            <rsc_ticket id="C3" ticket="T1" rsc="R1" rsc-role="Master" />
            <rsc_ticket id="C4" ticket="T1" rsc="R1" rsc-role="Promoted" />
            <rsc_ticket id="C5" ticket="T1" rsc="R1" rsc-role="Unpromoted" />
            <rsc_ticket id="C6" ticket="T2" rsc="R1" />
            <rsc_ticket id="C7" ticket="T1" rsc="R2" />
            <rsc_ticket id="C8" ticket="T1">
                <resource_set id="C8_set">
                    <resource_ref id="R1" />
                </resource_set>
            </rsc_ticket>
        </constraints>
        """
    )

    def test_success(self):
        duplicates = {
            "C1": ["C2"],  # literally same constraints
            "C2": ["C1"],  # literally same constraints
            "C3": ["C4"],  # same constraints, legacy role vs the same new role
            "C4": ["C3"],  # same constraints, legacy role vs the same new role
            "C5": [],  # role doesn't match
            "C6": [],  # ticket doesn't match
            "C7": [],  # resource doesn't match
        }
        checker = ticket.DuplicatesCheckerTicketPlain()
        self.assert_success(self.cib, checker, duplicates)


class DuplicatesCheckerTicketWithSet(DuplicatesCheckerTestBase):
    cib = etree.fromstring(
        """
        <constraints>
            <rsc_ticket id="C1" ticket="T1">
                <resource_set id="C1_set">
                    <resource_ref id="R1" /> <resource_ref id="R2" />
                </resource_set>
                <resource_set id="C1_set-1">
                    <resource_ref id="R3" /> <resource_ref id="R4" />
                </resource_set>
            </rsc_ticket>
            <rsc_ticket id="C2" ticket="T1">
                <resource_set id="C2_set" role="Promoted">
                    <resource_ref id="R1" /> <resource_ref id="R2" />
                </resource_set>
                <resource_set id="C2_set-1" role="Unpromoted">
                    <resource_ref id="R3" /> <resource_ref id="R4" />
                </resource_set>
            </rsc_ticket>
            <rsc_ticket id="C3" ticket="T2">
                <resource_set id="C3_set">
                    <resource_ref id="R1" /> <resource_ref id="R2" />
                </resource_set>
                <resource_set id="C3_set-1">
                    <resource_ref id="R3" /> <resource_ref id="R4" />
                </resource_set>
            </rsc_ticket>
            <rsc_ticket id="C4" ticket="T1">
                <resource_set id="C4_set">
                    <resource_ref id="R1" /> <resource_ref id="R2" />
                </resource_set>
            </rsc_ticket>
        </constraints>
        """
    )

    def test_success(self):
        duplicates = {
            "C1": ["C2"],  # same constraints, options don't matter
            "C2": ["C1"],  # same constraints, options don't matter
            "C3": [],  # same resources, different ticket
            "C4": [],  # different resources, same ticket
        }
        checker = ticket.DuplicatesCheckerTicketWithSet()
        self.assert_success(self.cib, checker, duplicates)


class RemovePlainTest(TestCase):
    def test_remove_tickets_constraints_for_resource(self):
        constraint_section = etree.fromstring(
            """
            <constraints>
                <rsc_ticket id="t1" ticket="tA" rsc="rA"/>
                <rsc_ticket id="t2" ticket="tA" rsc="rB"/>
                <rsc_ticket id="t3" ticket="tA" rsc="rA"/>
                <rsc_ticket id="t4" ticket="tB" rsc="rA"/>
                <rsc_ticket id="t5" ticket="tB" rsc="rB"/>
            </constraints>
        """
        )

        self.assertTrue(
            ticket.remove_plain(
                constraint_section,
                ticket_key="tA",
                resource_id="rA",
            )
        )

        assert_xml_equal(
            etree.tostring(constraint_section).decode(),
            """
            <constraints>
                <rsc_ticket id="t2" ticket="tA" rsc="rB"/>
                <rsc_ticket id="t4" ticket="tB" rsc="rA"/>
                <rsc_ticket id="t5" ticket="tB" rsc="rB"/>
            </constraints>
        """,
        )

    def test_remove_nothing_when_no_matching_found(self):
        constraint_section = etree.fromstring(
            """
            <constraints>
                <rsc_ticket id="t2" ticket="tA" rsc="rB"/>
                <rsc_ticket id="t4" ticket="tB" rsc="rA"/>
                <rsc_ticket id="t5" ticket="tB" rsc="rB"/>
            </constraints>
        """
        )

        self.assertFalse(
            ticket.remove_plain(
                constraint_section,
                ticket_key="tA",
                resource_id="rA",
            )
        )

        assert_xml_equal(
            etree.tostring(constraint_section).decode(),
            """
            <constraints>
                <rsc_ticket id="t2" ticket="tA" rsc="rB"/>
                <rsc_ticket id="t4" ticket="tB" rsc="rA"/>
                <rsc_ticket id="t5" ticket="tB" rsc="rB"/>
            </constraints>
        """,
        )


class RemoveWithSetTest(TestCase):
    def test_remove_resource_references_and_empty_remaining_parents(self):
        constraint_section = etree.fromstring(
            """
            <constraints>
                <rsc_ticket id="t1" ticket="tA">
                    <resource_set id="rs1">
                        <resource_ref id="rA"/>
                    </resource_set>
                    <resource_set id="rs2">
                        <resource_ref id="rA"/>
                    </resource_set>
                </rsc_ticket>

                <rsc_ticket id="t2" ticket="tA">
                    <resource_set id="rs3">
                        <resource_ref id="rA"/>
                        <resource_ref id="rB"/>
                    </resource_set>
                    <resource_set id="rs4">
                        <resource_ref id="rA"/>
                    </resource_set>
                </rsc_ticket>

                <rsc_ticket id="t3" ticket="tB">
                    <resource_set id="rs5">
                        <resource_ref id="rA"/>
                    </resource_set>
                </rsc_ticket>
            </constraints>
        """
        )

        self.assertTrue(
            ticket.remove_with_resource_set(
                constraint_section, ticket_key="tA", resource_id="rA"
            )
        )

        assert_xml_equal(
            """
                <constraints>
                    <rsc_ticket id="t2" ticket="tA">
                        <resource_set id="rs3">
                            <resource_ref id="rB"/>
                        </resource_set>
                    </rsc_ticket>

                    <rsc_ticket id="t3" ticket="tB">
                        <resource_set id="rs5">
                            <resource_ref id="rA"/>
                        </resource_set>
                    </rsc_ticket>
                </constraints>
            """,
            etree.tostring(constraint_section).decode(),
        )

    def test_remove_nothing_when_no_matching_found(self):
        constraint_section = etree.fromstring(
            """
                <constraints>
                    <rsc_ticket id="t2" ticket="tA">
                        <resource_set id="rs3">
                            <resource_ref id="rB"/>
                        </resource_set>
                    </rsc_ticket>

                    <rsc_ticket id="t3" ticket="tB">
                        <resource_set id="rs5">
                            <resource_ref id="rA"/>
                        </resource_set>
                    </rsc_ticket>
                </constraints>
        """
        )
        self.assertFalse(
            ticket.remove_with_resource_set(
                constraint_section, ticket_key="tA", resource_id="rA"
            )
        )



from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs.common import const
from pcs.common.reports import ReportItemSeverity as severities
from pcs.common.reports import codes as report_codes
from pcs.common.types import CibRuleInEffectStatus
from pcs.lib.commands.constraint import common as constraint

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_xml_equal,
)
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.constraints_dto import get_all_constraints
from pcs_test.tools.custom_mock import (
    MockLibraryReportProcessor,
    RuleInEffectEvalMock,
)


def fixture_cib_and_constraints():
    cib = etree.Element(
        "cib",
        {"validate-with": f"pacemaker-{const.PCMK_NEW_ROLES_CIB_VERSION}"},
    )
    config_section = etree.SubElement(cib, "configuration")
    resources_section = etree.SubElement(config_section, "resources")
    for _id in ("A", "B", "E", "F"):
        etree.SubElement(resources_section, "primitive").attrib["id"] = _id
    constraint_section = etree.SubElement(config_section, "constraints")
    return cib, constraint_section


def fixture_env(cib):
    env = mock.MagicMock()
    env.get_cib = mock.Mock()
    env.get_cib.return_value = cib
    env.push_cib = mock.Mock()
    env.report_processor = MockLibraryReportProcessor()
    return env


class CreateWithSetTest(TestCase):
    def setUp(self):
        self.cib, self.constraint_section = fixture_cib_and_constraints()
        self.env = fixture_env(self.cib)
        self.role = const.PCMK_ROLE_PROMOTED
        self.independent_cib = etree.XML(etree.tostring(self.cib))

    def create(self, duplication_allowed=False):
        constraint.create_with_set(
            "rsc_some",
            lambda cib, options, resource_set_list: options,
            self.env,
            [
                {"ids": ["A", "B"], "options": {"role": self.role}},
                {"ids": ["E", "F"], "options": {"action": "start"}},
            ],
            {"id": "some_id", "symmetrical": "true"},
            duplication_alowed=duplication_allowed,
        )

    def test_put_new_constraint_to_constraint_section(self):
        self.create()
        self.env.push_cib.assert_called_once_with()
        self.independent_cib.find(".//constraints").append(
            etree.XML(
                f"""
            <rsc_some id="some_id" symmetrical="true">
                  <resource_set id="some_id_set" role="{const.PCMK_ROLE_PROMOTED}">
                      <resource_ref id="A"></resource_ref>
                      <resource_ref id="B"></resource_ref>
                  </resource_set>
                  <resource_set action="start" id="some_id_set-1">
                      <resource_ref id="E"></resource_ref>
                      <resource_ref id="F"></resource_ref>
                  </resource_set>
            </rsc_some>
        """
            )
        )
        assert_xml_equal(
            etree.tostring(self.independent_cib).decode(),
            etree.tostring(self.cib).decode(),
        )

    def test_refuse_legacy_role(self):
        self.role = const.PCMK_ROLE_PROMOTED_LEGACY
        assert_raise_library_error(self.create)
        self.env.report_processor.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="role",
                    option_value=self.role,
                    allowed_values=const.PCMK_ROLES,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_refuse_duplicate(self):
        self.create()
        self.env.push_cib.assert_called_once_with()
        assert_raise_library_error(self.create)
        self.env.report_processor.assert_reports(
            [
                (
                    severities.ERROR,
                    report_codes.DUPLICATE_CONSTRAINTS_EXIST,
                    {
                        "constraint_ids": ["some_id"],
                    },
                    report_codes.FORCE,
                ),
            ]
        )

    def test_put_duplicate_constraint_when_duplication_allowed(self):
        self.create()
        self.create(duplication_allowed=True)
        expected_calls = [
            mock.call(),
            mock.call(),
        ]
        self.assertEqual(self.env.push_cib.call_count, len(expected_calls))
        self.env.push_cib.assert_has_calls(expected_calls)

        constraint_section = self.independent_cib.find(".//constraints")
        constraint_section.append(
            etree.XML(
                f"""
            <rsc_some id="some_id" symmetrical="true">
                <resource_set id="some_id_set" role="{const.PCMK_ROLE_PROMOTED}">
                    <resource_ref id="A"></resource_ref>
                    <resource_ref id="B"></resource_ref>
                </resource_set>
                <resource_set action="start" id="some_id_set-1">
                    <resource_ref id="E"></resource_ref>
                    <resource_ref id="F"></resource_ref>
                </resource_set>
            </rsc_some>
        """
            )
        )
        constraint_section.append(
            etree.XML(
                f"""
            <rsc_some id="some_id" symmetrical="true">
                <resource_set id="some_id_set-2" role="{const.PCMK_ROLE_PROMOTED}">
                    <resource_ref id="A"></resource_ref>
                    <resource_ref id="B"></resource_ref>
                </resource_set>
                <resource_set action="start" id="some_id_set-3">
                    <resource_ref id="E"></resource_ref>
                    <resource_ref id="F"></resource_ref>
                </resource_set>
            </rsc_some>
        """
            )
        )
        assert_xml_equal(
            etree.tostring(self.independent_cib).decode(),
            etree.tostring(self.cib).decode(),
        )


class GetConfig(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    @mock.patch("pcs.lib.commands.constraint.common.get_rule_evaluator")
    def test_success(self, get_rule_evaluator_mock):
        rule_evaluator = RuleInEffectEvalMock(
            {
                "loc_constr_with_expired_rule-rule": CibRuleInEffectStatus.EXPIRED,
                "loc_constr_with_not_expired_rule": CibRuleInEffectStatus.IN_EFFECT,
            }
        )
        self.config.runner.cib.load(filename="cib-all.xml")
        get_rule_evaluator_mock.return_value = rule_evaluator
        self.assertEqual(
            get_all_constraints(rule_evaluator),
            constraint.get_config(
                self.env_assist.get_env(), evaluate_rules=True
            ),
        )
        get_rule_evaluator_mock.assert_called_once()

    @mock.patch("pcs.lib.cib.rule.in_effect.has_rule_in_effect_status_tool")
    def test_no_rule_evaluation(self, has_rule_tool_mock):
        has_rule_tool_mock.side_effect = AssertionError(
            "has_rule_in_effect_status_tool should not be called"
        )
        self.config.runner.cib.load(filename="cib-all.xml")
        self.assertEqual(
            get_all_constraints(RuleInEffectEvalMock({})),
            constraint.get_config(
                self.env_assist.get_env(), evaluate_rules=False
            ),
        )



from pcs import (
    qdevice,
    usage,
)
from pcs.cli.common.routing import create_router

qdevice_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: print(usage.qdevice(argv)),
        "status": qdevice.qdevice_status_cmd,
        "setup": qdevice.qdevice_setup_cmd,
        "destroy": qdevice.qdevice_destroy_cmd,
        "start": qdevice.qdevice_start_cmd,
        "stop": qdevice.qdevice_stop_cmd,
        "kill": qdevice.qdevice_kill_cmd,
        "enable": qdevice.qdevice_enable_cmd,
        "disable": qdevice.qdevice_disable_cmd,
    },
    ["qdevice"],
)



import sys
import xml.dom.minidom
from enum import Enum
from typing import (
    Any,
    Iterable,
    Optional,
    Set,
    TypeVar,
    cast,
)
from xml.dom.minidom import parseString

import pcs.cli.constraint_order.command as order_command
from pcs import utils
from pcs.cli.common import parse_args
from pcs.cli.common.errors import (
    CmdLineInputError,
    raise_command_replaced,
)
from pcs.cli.common.output import (
    INDENT_STEP,
    lines_to_str,
)
from pcs.cli.constraint.location.command import (
    RESOURCE_TYPE_REGEXP,
    RESOURCE_TYPE_RESOURCE,
)
from pcs.cli.constraint.output import (
    CibConstraintLocationAnyDto,
    filter_constraints_by_rule_expired_status,
    location,
    print_config,
)
from pcs.cli.reports import process_library_reports
from pcs.cli.reports.output import (
    deprecation_warning,
    print_to_stderr,
    warn,
)
from pcs.common import (
    const,
    pacemaker,
    reports,
)
from pcs.common.pacemaker.constraint import (
    CibConstraintColocationSetDto,
    CibConstraintLocationSetDto,
    CibConstraintOrderSetDto,
    CibConstraintsDto,
    CibConstraintTicketSetDto,
    get_all_constraints_ids,
)
from pcs.common.pacemaker.resource.list import CibResourcesDto
from pcs.common.pacemaker.types import CibResourceDiscovery
from pcs.common.reports import ReportItem
from pcs.common.str_tools import (
    format_list,
    indent,
)
from pcs.common.types import (
    StringCollection,
    StringIterable,
    StringSequence,
)
from pcs.lib.cib.constraint.order import ATTRIB as order_attrib
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.pacemaker.values import (
    SCORE_INFINITY,
    is_true,
    sanitize_id,
)

# pylint: disable=invalid-name
# pylint: disable=too-many-branches
# pylint: disable=too-many-lines
# pylint: disable=too-many-locals
# pylint: disable=too-many-statements

DEFAULT_ACTION = const.PCMK_ACTION_START
DEFAULT_ROLE = const.PCMK_ROLE_STARTED

OPTIONS_SYMMETRICAL = order_attrib["symmetrical"]

LOCATION_NODE_VALIDATION_SKIP_MSG = (
    "Validation for node existence in the cluster will be skipped"
)
STANDALONE_SCORE_MSG = (
    "Specifying score as a standalone value is deprecated and "
    "might be removed in a future release, use score=value instead"
)


class CrmRuleReturnCode(Enum):
    IN_EFFECT = 0
    EXPIRED = 110
    TO_BE_IN_EFFECT = 111


def constraint_order_cmd(lib, argv, modifiers):
    sub_cmd = "config" if not argv else argv.pop(0)

    try:
        if sub_cmd == "set":
            order_command.create_with_set(lib, argv, modifiers)
        elif sub_cmd in ["remove", "delete"]:
            order_rm(lib, argv, modifiers)
        elif sub_cmd == "show":
            raise_command_replaced(
                ["pcs constraint order config"], pcs_version="0.12"
            )
        elif sub_cmd == "config":
            order_command.config_cmd(lib, argv, modifiers)
        else:
            order_start(lib, [sub_cmd] + argv, modifiers)
    except CmdLineInputError as e:
        utils.exit_on_cmdline_input_error(e, "constraint", ["order", sub_cmd])


def config_cmd(
    lib: Any, argv: list[str], modifiers: parse_args.InputModifiers
) -> None:
    modifiers.ensure_only_supported("-f", "--output-format", "--full", "--all")
    if argv:
        raise CmdLineInputError()

    print_config(
        cast(
            CibConstraintsDto,
            lib.constraint.get_config(evaluate_rules=True),
        ),
        modifiers,
    )


def _validate_constraint_resource(cib_dom, resource_id):
    (
        resource_valid,
        resource_error,
        dummy_correct_id,
    ) = utils.validate_constraint_resource(cib_dom, resource_id)
    if not resource_valid:
        utils.err(resource_error)


def _validate_resources_not_in_same_group(cib_dom, resource1, resource2):
    if not utils.validate_resources_not_in_same_group(
        cib_dom, resource1, resource2
    ):
        utils.err(
            "Cannot create an order constraint for resources in the same group"
        )


# Syntax: colocation add [role] <src> with [role] <tgt> [score] [options]
# possible commands:
#        <src> with        <tgt> [score] [options]
#        <src> with <role> <tgt> [score] [options]
# <role> <src> with        <tgt> [score] [options]
# <role> <src> with <role> <tgt> [score] [options]
# Specifying score as a single argument is deprecated, though. The correct way
# is score=value in options.
def colocation_add(lib, argv, modifiers):  # noqa: PLR0912, PLR0915
    """
    Options:
      * -f - CIB file
      * --force - allow constraint on any resource, allow duplicate constraints
    """

    def _parse_score_options(argv):
        # When passed an array of arguments if the first argument doesn't have
        # an '=' then it's the score, otherwise they're all arguments. Return a
        # tuple with the score and array of name,value pairs
        """
        Commandline options: no options
        """
        if not argv:
            return None, []
        score = None
        if "=" not in argv[0]:
            score = argv.pop(0)
            # TODO added to pcs in the first 0.12.x version
            deprecation_warning(STANDALONE_SCORE_MSG)

        # create a list of 2-tuples (name, value)
        arg_array = [
            parse_args.split_option(arg, allow_empty_value=False)
            for arg in argv
        ]
        return score, arg_array

    del lib
    modifiers.ensure_only_supported("-f", "--force")
    if len(argv) < 3:
        raise CmdLineInputError()

    role1 = ""
    role2 = ""

    cib_dom = utils.get_cib_dom()
    new_roles_supported = utils.isCibVersionSatisfied(
        cib_dom, const.PCMK_NEW_ROLES_CIB_VERSION
    )

    def _validate_and_prepare_role(role):
        role_cleaned = role.lower().capitalize()
        if role_cleaned not in const.PCMK_ROLES:
            utils.err(
                "invalid role value '{0}', allowed values are: {1}".format(
                    role, format_list(const.PCMK_ROLES)
                )
            )
        return pacemaker.role.get_value_for_cib(
            role_cleaned, new_roles_supported
        )

    if argv[2] == "with":
        role1 = _validate_and_prepare_role(argv.pop(0))
        resource1 = argv.pop(0)
    elif argv[1] == "with":
        resource1 = argv.pop(0)
    else:
        raise CmdLineInputError()

    if argv.pop(0) != "with":
        raise CmdLineInputError()
    if "with" in argv:
        raise CmdLineInputError(
            message="Multiple 'with's cannot be specified.",
            hint=(
                "Use the 'pcs constraint colocation set' command if you want "
                "to create a constraint for more than two resources."
            ),
            show_both_usage_and_message=True,
        )

    if not argv:
        raise CmdLineInputError()
    if len(argv) == 1 or utils.is_score_or_opt(argv[1]):
        resource2 = argv.pop(0)
    else:
        role2 = _validate_and_prepare_role(argv.pop(0))
        resource2 = argv.pop(0)

    score, nv_pairs = _parse_score_options(argv)

    _validate_constraint_resource(cib_dom, resource1)
    _validate_constraint_resource(cib_dom, resource2)

    id_in_nvpairs = None
    for name, value in nv_pairs:
        if name == "id":
            id_valid, id_error = utils.validate_xml_id(value, "constraint id")
            if not id_valid:
                utils.err(id_error)
            if utils.does_id_exist(cib_dom, value):
                utils.err(
                    "id '%s' is already in use, please specify another one"
                    % value
                )
            id_in_nvpairs = True
        elif name == "score":
            score = value
    if score is None:
        score = SCORE_INFINITY
    if not id_in_nvpairs:
        nv_pairs.append(
            (
                "id",
                utils.find_unique_id(
                    cib_dom,
                    "colocation-%s-%s-%s" % (resource1, resource2, score),
                ),
            )
        )

    (dom, constraintsElement) = getCurrentConstraints(cib_dom)

    # If one role is specified, the other should default to "started"
    if role1 != "" and role2 == "":
        role2 = DEFAULT_ROLE
    if role2 != "" and role1 == "":
        role1 = DEFAULT_ROLE
    element = dom.createElement("rsc_colocation")
    element.setAttribute("rsc", resource1)
    element.setAttribute("with-rsc", resource2)
    element.setAttribute("score", score)
    if role1 != "":
        element.setAttribute("rsc-role", role1)
    if role2 != "":
        element.setAttribute("with-rsc-role", role2)
    for nv_pair in nv_pairs:
        element.setAttribute(nv_pair[0], nv_pair[1])
    if not modifiers.get("--force"):

        def _constraint_export(constraint_info):
            options_dict = constraint_info["options"]
            co_resource1 = options_dict.get("rsc", "")
            co_resource2 = options_dict.get("with-rsc", "")
            co_id = options_dict.get("id", "")
            co_score = options_dict.get("score", "")
            score_text = "(score:" + co_score + ")"
            console_option_list = [
                f"({option[0]}:{option[1]})"
                for option in sorted(options_dict.items())
                if option[0] not in ("rsc", "with-rsc", "id", "score")
            ]
            console_option_list.append(f"(id:{co_id})")
            return " ".join(
                [co_resource1, "with", co_resource2, score_text]
                + console_option_list
            )

        duplicates = colocation_find_duplicates(constraintsElement, element)
        if duplicates:
            utils.err(
                "duplicate constraint already exists, use --force to override\n"
                + "\n".join(
                    [
                        "  "
                        + _constraint_export(
                            {"options": dict(dup.attributes.items())}
                        )
                        for dup in duplicates
                    ]
                )
            )
    constraintsElement.appendChild(element)
    utils.replace_cib_configuration(dom)


def colocation_find_duplicates(dom, constraint_el):
    """
    Commandline options: no options
    """
    new_roles_supported = utils.isCibVersionSatisfied(
        dom, const.PCMK_NEW_ROLES_CIB_VERSION
    )

    def normalize(const_el):
        return (
            const_el.getAttribute("rsc"),
            const_el.getAttribute("with-rsc"),
            pacemaker.role.get_value_for_cib(
                const_el.getAttribute("rsc-role").capitalize() or DEFAULT_ROLE,
                new_roles_supported,
            ),
            pacemaker.role.get_value_for_cib(
                const_el.getAttribute("with-rsc-role").capitalize()
                or DEFAULT_ROLE,
                new_roles_supported,
            ),
        )

    normalized_el = normalize(constraint_el)
    return [
        other_el
        for other_el in dom.getElementsByTagName("rsc_colocation")
        if not other_el.getElementsByTagName("resource_set")
        and constraint_el is not other_el
        and normalized_el == normalize(other_el)
    ]


def order_rm(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
    """
    del lib
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    elementFound = False
    (dom, constraintsElement) = getCurrentConstraints()

    for resource in argv:
        for ord_loc in constraintsElement.getElementsByTagName("rsc_order")[:]:
            if (
                ord_loc.getAttribute("first") == resource
                or ord_loc.getAttribute("then") == resource
            ):
                constraintsElement.removeChild(ord_loc)
                elementFound = True

        resource_refs_to_remove = []
        for ord_set in constraintsElement.getElementsByTagName("resource_ref"):
            if ord_set.getAttribute("id") == resource:
                resource_refs_to_remove.append(ord_set)
                elementFound = True

        for res_ref in resource_refs_to_remove:
            res_set = res_ref.parentNode
            res_order = res_set.parentNode

            res_ref.parentNode.removeChild(res_ref)
            if not res_set.getElementsByTagName("resource_ref"):
                res_set.parentNode.removeChild(res_set)
                if not res_order.getElementsByTagName("resource_set"):
                    res_order.parentNode.removeChild(res_order)

    if elementFound:
        utils.replace_cib_configuration(dom)
    else:
        utils.err("No matching resources found in ordering list")


def order_start(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
      * --force - allow constraint for any resource, allow duplicate constraints
    """
    del lib
    modifiers.ensure_only_supported("-f", "--force")
    if len(argv) < 3:
        raise CmdLineInputError()

    first_action = DEFAULT_ACTION
    then_action = DEFAULT_ACTION
    action = argv[0]
    if action in const.PCMK_ACTIONS:
        first_action = action
        argv.pop(0)

    resource1 = argv.pop(0)
    if argv.pop(0) != "then":
        raise CmdLineInputError()

    if not argv:
        raise CmdLineInputError()

    action = argv[0]
    if action in const.PCMK_ACTIONS:
        then_action = action
        argv.pop(0)

    if not argv:
        raise CmdLineInputError()
    resource2 = argv.pop(0)

    order_options = []
    if argv:
        order_options = order_options + argv[:]
    if "then" in order_options:
        raise CmdLineInputError(
            message="Multiple 'then's cannot be specified.",
            hint=(
                "Use the 'pcs constraint order set' command if you want to "
                "create a constraint for more than two resources."
            ),
            show_both_usage_and_message=True,
        )

    order_options.append("first-action=" + first_action)
    order_options.append("then-action=" + then_action)
    _order_add(resource1, resource2, order_options, modifiers)


def _order_add(resource1, resource2, options_list, modifiers):  # noqa: PLR0912, PLR0915
    """
    Commandline options:
      * -f - CIB file
      * --force - allow constraint for any resource, allow duplicate constraints
    """
    cib_dom = utils.get_cib_dom()
    _validate_constraint_resource(cib_dom, resource1)
    _validate_constraint_resource(cib_dom, resource2)

    _validate_resources_not_in_same_group(cib_dom, resource1, resource2)

    order_options = []
    id_specified = False
    sym = None
    for arg in options_list:
        if arg == "symmetrical":
            sym = "true"
        elif arg == "nonsymmetrical":
            sym = "false"
        else:
            name, value = parse_args.split_option(arg, allow_empty_value=False)
            if name == "id":
                id_valid, id_error = utils.validate_xml_id(
                    value, "constraint id"
                )
                if not id_valid:
                    utils.err(id_error)
                if utils.does_id_exist(cib_dom, value):
                    utils.err(
                        "id '%s' is already in use, please specify another one"
                        % value
                    )
                id_specified = True
                order_options.append((name, value))
            elif name == "symmetrical":
                if value.lower() in OPTIONS_SYMMETRICAL:
                    sym = value.lower()
                else:
                    utils.err(
                        "invalid symmetrical value '%s', allowed values are: %s"
                        % (value, ", ".join(OPTIONS_SYMMETRICAL))
                    )
            else:
                order_options.append((name, value))
    if sym:
        order_options.append(("symmetrical", sym))

    options = ""
    if order_options:
        options = " (Options: %s)" % " ".join(
            [
                "%s=%s" % (name, value)
                for name, value in order_options
                if name not in ("kind", "score")
            ]
        )

    scorekind = "kind: Mandatory"
    id_suffix = "mandatory"
    for opt in order_options:
        if opt[0] == "score":
            scorekind = "score: " + opt[1]
            id_suffix = opt[1]
            # TODO deprecated in pacemaker 2, to be removed in pacemaker 3
            # added to pcs after 0.11.7
            deprecation_warning(
                reports.messages.DeprecatedOption(opt[0], []).message
            )
            break
        if opt[0] == "kind":
            scorekind = "kind: " + opt[1]
            id_suffix = opt[1]
            break

    if not id_specified:
        order_id = "order-" + resource1 + "-" + resource2 + "-" + id_suffix
        order_id = utils.find_unique_id(cib_dom, order_id)
        order_options.append(("id", order_id))

    (dom, constraintsElement) = getCurrentConstraints()
    element = dom.createElement("rsc_order")
    element.setAttribute("first", resource1)
    element.setAttribute("then", resource2)
    for order_opt in order_options:
        element.setAttribute(order_opt[0], order_opt[1])
    constraintsElement.appendChild(element)
    if not modifiers.get("--force"):

        def _constraint_export(constraint_info):
            options = constraint_info["options"]
            oc_resource1 = options.get("first", "")
            oc_resource2 = options.get("then", "")
            first_action = options.get("first-action", "")
            then_action = options.get("then-action", "")
            oc_id = options.get("id", "")
            oc_score = options.get("score", "")
            oc_kind = options.get("kind", "")
            oc_sym = ""
            oc_id_out = ""
            oc_options = ""
            if "symmetrical" in options and not is_true(
                options.get("symmetrical", "false")
            ):
                oc_sym = "(non-symmetrical)"
            if oc_kind != "":
                score_text = "(kind:" + oc_kind + ")"
            elif oc_kind == "" and oc_score == "":
                score_text = "(kind:Mandatory)"
            else:
                score_text = "(score:" + oc_score + ")"
            oc_id_out = "(id:" + oc_id + ")"
            already_processed_options = (
                "first",
                "then",
                "first-action",
                "then-action",
                "id",
                "score",
                "kind",
                "symmetrical",
            )
            oc_options = " ".join(
                [
                    f"{name}={value}"
                    for name, value in options.items()
                    if name not in already_processed_options
                ]
            )
            if oc_options:
                oc_options = "(Options: " + oc_options + ")"
            return " ".join(
                [
                    arg
                    for arg in [
                        first_action,
                        oc_resource1,
                        "then",
                        then_action,
                        oc_resource2,
                        score_text,
                        oc_sym,
                        oc_options,
                        oc_id_out,
                    ]
                    if arg
                ]
            )

        duplicates = order_find_duplicates(constraintsElement, element)
        if duplicates:
            utils.err(
                "duplicate constraint already exists, use --force to override\n"
                + "\n".join(
                    [
                        "  "
                        + _constraint_export(
                            {"options": dict(dup.attributes.items())}
                        )
                        for dup in duplicates
                    ]
                )
            )
    print_to_stderr(f"Adding {resource1} {resource2} ({scorekind}){options}")
    utils.replace_cib_configuration(dom)


def order_find_duplicates(dom, constraint_el):
    """
    Commandline options: no options
    """

    def normalize(constraint_el):
        return (
            constraint_el.getAttribute("first"),
            constraint_el.getAttribute("then"),
            constraint_el.getAttribute("first-action").lower()
            or DEFAULT_ACTION,
            constraint_el.getAttribute("then-action").lower() or DEFAULT_ACTION,
        )

    normalized_el = normalize(constraint_el)
    return [
        other_el
        for other_el in dom.getElementsByTagName("rsc_order")
        if not other_el.getElementsByTagName("resource_set")
        and constraint_el is not other_el
        and normalized_el == normalize(other_el)
    ]


_SetConstraint = TypeVar(
    "_SetConstraint",
    CibConstraintLocationSetDto,
    CibConstraintColocationSetDto,
    CibConstraintOrderSetDto,
    CibConstraintTicketSetDto,
)


def _filter_set_constraints_by_resources(
    constraints_dto: Iterable[_SetConstraint], resources: Set[str]
) -> list[_SetConstraint]:
    return [
        constraint_set_dto
        for constraint_set_dto in constraints_dto
        if any(
            set(resource_set.resources_ids) & resources
            for resource_set in constraint_set_dto.resource_sets
        )
    ]


def _filter_constraints_by_resources(
    constraints_dto: CibConstraintsDto,
    resources: StringIterable,
    patterns: StringIterable,
) -> CibConstraintsDto:
    required_resources_set = set(resources)
    required_patterns_set = set(patterns)
    return CibConstraintsDto(
        location=[
            constraint_dto
            for constraint_dto in constraints_dto.location
            if (
                constraint_dto.resource_id is not None
                and constraint_dto.resource_id in required_resources_set
            )
            or (
                constraint_dto.resource_pattern is not None
                and constraint_dto.resource_pattern in required_patterns_set
            )
        ],
        location_set=_filter_set_constraints_by_resources(
            constraints_dto.location_set, required_resources_set
        ),
        colocation=[
            constraint_dto
            for constraint_dto in constraints_dto.colocation
            if {constraint_dto.resource_id, constraint_dto.with_resource_id}
            & required_resources_set
        ],
        colocation_set=_filter_set_constraints_by_resources(
            constraints_dto.colocation_set, required_resources_set
        ),
        order=[
            constraint_dto
            for constraint_dto in constraints_dto.order
            if {
                constraint_dto.first_resource_id,
                constraint_dto.then_resource_id,
            }
            & required_resources_set
        ],
        order_set=_filter_set_constraints_by_resources(
            constraints_dto.order_set, required_resources_set
        ),
        ticket=[
            constraint_dto
            for constraint_dto in constraints_dto.ticket
            if constraint_dto.resource_id in required_resources_set
        ],
        ticket_set=_filter_set_constraints_by_resources(
            constraints_dto.ticket_set, required_resources_set
        ),
    )


def _filter_location_by_node_base(
    constraint_dtos: Iterable[CibConstraintLocationAnyDto],
    nodes: StringCollection,
) -> list[CibConstraintLocationAnyDto]:
    return [
        constraint_dto
        for constraint_dto in constraint_dtos
        if constraint_dto.attributes.node is not None
        and constraint_dto.attributes.node in nodes
    ]


def location_config_cmd(
    lib: Any, argv: parse_args.Argv, modifiers: parse_args.InputModifiers
) -> None:
    """
    Options:
      * --all - print expired constraints
      * --full - print all details
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f", "--output-format", "--full", "--all")
    filter_type: Optional[str] = None
    filter_items: parse_args.Argv = []
    if argv:
        filter_type, *filter_items = argv
        allowed_types = ("resources", "nodes")
        if filter_type not in allowed_types:
            raise CmdLineInputError(
                f"Unknown keyword '{filter_type}'. Allowed keywords: "
                f"{format_list(allowed_types)}"
            )
        if modifiers.get_output_format() != parse_args.OUTPUT_FORMAT_VALUE_TEXT:
            raise CmdLineInputError(
                "Output formats other than 'text' are not supported together "
                "with grouping and filtering by nodes or resources"
            )

    constraints_dto = filter_constraints_by_rule_expired_status(
        lib.constraint.get_config(evaluate_rules=True),
        modifiers.is_specified("--all"),
    )

    constraints_dto = CibConstraintsDto(
        location=constraints_dto.location,
        location_set=constraints_dto.location_set,
    )

    def _print_lines(lines: StringSequence) -> None:
        if lines:
            print("Location Constraints:")
            print(lines_to_str(indent(lines, indent_step=INDENT_STEP)))

    if filter_type == "resources":
        if filter_items:
            resources = []
            patterns = []
            for item in filter_items:
                item_type, item_value = parse_args.parse_typed_arg(
                    item,
                    [RESOURCE_TYPE_RESOURCE, RESOURCE_TYPE_REGEXP],
                    RESOURCE_TYPE_RESOURCE,
                )
                if item_type == RESOURCE_TYPE_RESOURCE:
                    resources.append(item_value)
                elif item_type == RESOURCE_TYPE_REGEXP:
                    patterns.append(item_value)
            constraints_dto = _filter_constraints_by_resources(
                constraints_dto, resources, patterns
            )
        _print_lines(
            location.constraints_to_grouped_by_resource_text(
                constraints_dto.location,
                modifiers.is_specified("--full"),
            )
        )
        return
    if filter_type == "nodes":
        if filter_items:
            constraints_dto = CibConstraintsDto(
                location=_filter_location_by_node_base(
                    constraints_dto.location, filter_items
                ),
                location_set=_filter_location_by_node_base(
                    constraints_dto.location_set, filter_items
                ),
            )
        _print_lines(
            location.constraints_to_grouped_by_node_text(
                constraints_dto.location,
                modifiers.is_specified("--full"),
            )
        )
        return

    print_config(constraints_dto, modifiers)


def _verify_node_name(node, existing_nodes):
    report_list = []
    if node not in existing_nodes:
        report_list.append(
            ReportItem.error(
                reports.messages.NodeNotFound(node),
                force_code=reports.codes.FORCE,
            )
        )
    return report_list


def _verify_score(score):
    if not utils.is_score(score):
        utils.err(
            "invalid score '%s', use integer or INFINITY or -INFINITY" % score
        )


def location_prefer(  # noqa: PLR0912
    lib: Any, argv: parse_args.Argv, modifiers: parse_args.InputModifiers
) -> None:
    """
    Options:
      * --force - allow unknown options, allow constraint for any resource type
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "-f")
    rsc = argv.pop(0)
    prefer_option = argv.pop(0)

    dummy_rsc_type, rsc_value = parse_args.parse_typed_arg(
        rsc,
        [RESOURCE_TYPE_RESOURCE, RESOURCE_TYPE_REGEXP],
        RESOURCE_TYPE_RESOURCE,
    )

    if prefer_option == "prefers":
        prefer = True
    elif prefer_option == "avoids":
        prefer = False
    else:
        raise CmdLineInputError()

    skip_node_check = False
    existing_nodes: list[str] = []
    if modifiers.is_specified("-f") or modifiers.get("--force"):
        skip_node_check = True
        warn(LOCATION_NODE_VALIDATION_SKIP_MSG)
    else:
        lib_env = utils.get_lib_env()
        existing_nodes, report_list = get_existing_nodes_names(
            corosync_conf=lib_env.get_corosync_conf(),
            cib=lib_env.get_cib(),
        )
        if report_list:
            process_library_reports(report_list)

    report_list = []
    parameters_list = []
    for nodeconf in argv:
        nodeconf_a = nodeconf.split("=", 1)
        node = nodeconf_a[0]
        if not skip_node_check:
            report_list += _verify_node_name(node, existing_nodes)
        if len(nodeconf_a) == 1:
            score = "INFINITY" if prefer else "-INFINITY"
        else:
            score = nodeconf_a[1]
            _verify_score(score)
            if not prefer:
                score = score[1:] if score[0] == "-" else "-" + score

        parameters_list.append(
            [
                sanitize_id(f"location-{rsc_value}-{node}-{score}"),
                rsc,
                node,
                f"score={score}",
            ]
        )

    if report_list:
        process_library_reports(report_list)

    modifiers = modifiers.get_subset("--force", "-f")

    for parameters in parameters_list:
        location_add(lib, parameters, modifiers, skip_score_and_node_check=True)


def location_add(  # noqa: PLR0912, PLR0915
    lib: Any,
    argv: parse_args.Argv,
    modifiers: parse_args.InputModifiers,
    skip_score_and_node_check: bool = False,
) -> None:
    """
    Options:
      * --force - allow unknown options, allow constraint for any resource type
      * -f - CIB file
    """
    del lib
    modifiers.ensure_only_supported("--force", "-f")
    if len(argv) < 4:
        raise CmdLineInputError()

    constraint_id = argv.pop(0)
    rsc_type, rsc_value = parse_args.parse_typed_arg(
        argv.pop(0),
        [RESOURCE_TYPE_RESOURCE, RESOURCE_TYPE_REGEXP],
        RESOURCE_TYPE_RESOURCE,
    )
    node = argv.pop(0)
    score = None
    if "=" not in argv[0]:
        score = argv.pop(0)
        # TODO added to pcs in the first 0.12.x version
        deprecation_warning(STANDALONE_SCORE_MSG)
    options = []
    # For now we only allow setting resource-discovery and score
    for arg in argv:
        name, value = parse_args.split_option(arg, allow_empty_value=False)
        if name == "score":
            score = value
        elif name == "resource-discovery":
            if not modifiers.get("--force"):
                allowed_discovery = list(
                    map(
                        str,
                        [
                            CibResourceDiscovery.ALWAYS,
                            CibResourceDiscovery.EXCLUSIVE,
                            CibResourceDiscovery.NEVER,
                        ],
                    )
                )
                if value not in allowed_discovery:
                    utils.err(
                        (
                            "invalid {0} value '{1}', allowed values are: {2}"
                            ", use --force to override"
                        ).format(name, value, format_list(allowed_discovery))
                    )
            options.append([name, value])
        elif modifiers.get("--force"):
            options.append([name, value])
        else:
            utils.err("bad option '%s', use --force to override" % name)
    if score is None:
        score = "INFINITY"

    # Verify that specified node exists in the cluster and score is valid
    if not skip_score_and_node_check:
        if modifiers.is_specified("-f") or modifiers.get("--force"):
            warn(LOCATION_NODE_VALIDATION_SKIP_MSG)
        else:
            lib_env = utils.get_lib_env()
            existing_nodes, report_list = get_existing_nodes_names(
                corosync_conf=lib_env.get_corosync_conf(),
                cib=lib_env.get_cib(),
            )
            report_list += _verify_node_name(node, existing_nodes)
            if report_list:
                process_library_reports(report_list)
        _verify_score(score)

    id_valid, id_error = utils.validate_xml_id(constraint_id, "constraint id")
    if not id_valid:
        utils.err(id_error)

    dom = utils.get_cib_dom()

    if rsc_type == RESOURCE_TYPE_RESOURCE:
        (
            rsc_valid,
            rsc_error,
            dummy_correct_id,
        ) = utils.validate_constraint_resource(dom, rsc_value)
        if not rsc_valid:
            utils.err(rsc_error)

    # Verify current constraint doesn't already exist
    # If it does we replace it with the new constraint
    dummy_dom, constraintsElement = getCurrentConstraints(dom)
    # If the id matches, or the rsc & node match, then we replace/remove
    elementsToRemove = [
        rsc_loc
        for rsc_loc in constraintsElement.getElementsByTagName("rsc_location")
        # pylint: disable=too-many-boolean-expressions
        if rsc_loc.getAttribute("id") == constraint_id
        or (
            rsc_loc.getAttribute("node") == node
            and (
                (
                    rsc_type == RESOURCE_TYPE_RESOURCE
                    and rsc_loc.getAttribute("rsc") == rsc_value
                )
                or (
                    rsc_type == RESOURCE_TYPE_REGEXP
                    and rsc_loc.getAttribute("rsc-pattern") == rsc_value
                )
            )
        )
    ]
    for etr in elementsToRemove:
        constraintsElement.removeChild(etr)

    element = dom.createElement("rsc_location")
    element.setAttribute("id", constraint_id)
    if rsc_type == RESOURCE_TYPE_RESOURCE:
        element.setAttribute("rsc", rsc_value)
    elif rsc_type == RESOURCE_TYPE_REGEXP:
        element.setAttribute("rsc-pattern", rsc_value)
    element.setAttribute("node", node)
    element.setAttribute("score", score)
    for option in options:
        element.setAttribute(option[0], option[1])
    constraintsElement.appendChild(element)

    utils.replace_cib_configuration(dom)


# Grabs the current constraints and returns the dom and constraint element
def getCurrentConstraints(passed_dom=None):
    """
    Commandline options:
      * -f - CIB file, only if passed_dom is None
    """
    if passed_dom:
        dom = passed_dom
    else:
        current_constraints_xml = utils.get_cib_xpath("//constraints")
        if current_constraints_xml == "":
            utils.err("unable to process cib")
        # Verify current constraint doesn't already exist
        # If it does we replace it with the new constraint
        dom = parseString(current_constraints_xml)

    constraintsElement = dom.getElementsByTagName("constraints")[0]
    return (dom, constraintsElement)


# If returnStatus is set, then we don't error out, we just print the error
# and return false
def constraint_rm(  # noqa: PLR0912
    lib,
    argv,
    modifiers,
    returnStatus=False,
    constraintsElement=None,
    passed_dom=None,
):
    """
    Options:
      * -f - CIB file, effective only if passed_dom is None
    """
    if passed_dom is None:
        modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    bad_constraint = False
    if len(argv) != 1:
        for arg in argv:
            if not constraint_rm(
                lib, [arg], modifiers, returnStatus=True, passed_dom=passed_dom
            ):
                bad_constraint = True
        if bad_constraint:
            sys.exit(1)
        return None

    c_id = argv.pop(0)
    elementFound = False
    dom = None
    use_cibadmin = False
    if not constraintsElement:
        (dom, constraintsElement) = getCurrentConstraints(passed_dom)
        use_cibadmin = True

    for co in constraintsElement.childNodes[:]:
        if co.nodeType != xml.dom.Node.ELEMENT_NODE:
            continue
        if co.getAttribute("id") == c_id:
            constraintsElement.removeChild(co)
            elementFound = True

    if not elementFound:
        for rule in constraintsElement.getElementsByTagName("rule")[:]:
            if rule.getAttribute("id") == c_id:
                elementFound = True
                parent = rule.parentNode
                parent.removeChild(rule)
                if not parent.getElementsByTagName("rule"):
                    parent.parentNode.removeChild(parent)

    if elementFound:
        if passed_dom:
            return dom
        if use_cibadmin:
            utils.replace_cib_configuration(dom)
        if returnStatus:
            return True
    else:
        utils.err("Unable to find constraint - '%s'" % c_id, False)
        if returnStatus:
            return False
        sys.exit(1)
    return None


def _split_set_constraints(
    constraints_dto: CibConstraintsDto,
) -> tuple[CibConstraintsDto, CibConstraintsDto]:
    return (
        CibConstraintsDto(
            location=constraints_dto.location,
            colocation=constraints_dto.colocation,
            order=constraints_dto.order,
            ticket=constraints_dto.ticket,
        ),
        CibConstraintsDto(
            location_set=constraints_dto.location_set,
            colocation_set=constraints_dto.colocation_set,
            order_set=constraints_dto.order_set,
            ticket_set=constraints_dto.ticket_set,
        ),
    )


def _find_constraints_containing_resource(
    resources_dto: CibResourcesDto,
    constraints_dto: CibConstraintsDto,
    resource_id: str,
) -> CibConstraintsDto:
    resources_filter = [resource_id]
    # Original implementation only included parent resource only if resource_id
    # was referring to a primitive resource, ignoring groups. This may change in
    # the future if necessary.
    if any(
        primitive_dto.id == resource_id
        for primitive_dto in resources_dto.primitives
    ):
        for clone_dto in resources_dto.clones:
            if clone_dto.member_id == resource_id:
                resources_filter.append(clone_dto.id)
                break
    return _filter_constraints_by_resources(
        constraints_dto, resources_filter, []
    )


def ref(
    lib: Any, argv: list[str], modifiers: parse_args.InputModifiers
) -> None:
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    resources_dto = cast(
        CibResourcesDto, lib.resource.get_configured_resources()
    )

    constraints_dto = cast(
        CibConstraintsDto,
        lib.constraint.get_config(evaluate_rules=False),
    )

    for resource_id in sorted(set(argv)):
        constraint_ids = get_all_constraints_ids(
            _find_constraints_containing_resource(
                resources_dto, constraints_dto, resource_id
            )
        )
        print(f"Resource: {resource_id}")
        if constraint_ids:
            print(
                "\n".join(
                    indent(
                        sorted(constraint_ids),
                        indent_step=INDENT_STEP,
                    )
                )
            )
        else:
            print("  No Matches")


def remove_constraints_containing(
    resource_id: str, output=False, constraints_element=None, passed_dom=None
):
    """
    Commandline options:
      * -f - CIB file, effective only if passed_dom is None
    """
    lib = utils.get_library_wrapper()
    modifiers = utils.get_input_modifiers()
    resources_dto = cast(
        CibResourcesDto, lib.resource.get_configured_resources()
    )

    constraints_dto, set_constraints_dto = _split_set_constraints(
        cast(
            CibConstraintsDto,
            lib.constraint.get_config(evaluate_rules=False),
        )
    )
    constraints = sorted(
        get_all_constraints_ids(
            _find_constraints_containing_resource(
                resources_dto, constraints_dto, resource_id
            )
        )
    )
    set_constraints = sorted(
        get_all_constraints_ids(
            _find_constraints_containing_resource(
                resources_dto, set_constraints_dto, resource_id
            )
        )
    )
    for c in constraints:
        if output:
            print_to_stderr(f"Removing Constraint - {c}")
        if constraints_element is not None:
            constraint_rm(
                lib,
                [c],
                modifiers,
                True,
                constraints_element,
                passed_dom=passed_dom,
            )
        else:
            constraint_rm(lib, [c], modifiers, passed_dom=passed_dom)

    if set_constraints:
        (dom, constraintsElement) = getCurrentConstraints(passed_dom)
        for set_c in constraintsElement.getElementsByTagName("resource_ref")[:]:
            # If resource id is in a set, remove it from the set, if the set
            # is empty, then we remove the set, if the parent of the set
            # is empty then we remove it
            if set_c.getAttribute("id") == resource_id:
                parent_node = set_c.parentNode
                parent_node.removeChild(set_c)
                if output:
                    print_to_stderr(
                        "Removing {} from set {}".format(
                            resource_id, parent_node.getAttribute("id")
                        )
                    )
                if parent_node.getElementsByTagName("resource_ref").length == 0:
                    print_to_stderr(
                        "Removing set {}".format(parent_node.getAttribute("id"))
                    )
                    parent_node_2 = parent_node.parentNode
                    parent_node_2.removeChild(parent_node)
                    if (
                        parent_node_2.getElementsByTagName(
                            "resource_set"
                        ).length
                        == 0
                    ):
                        parent_node_2.parentNode.removeChild(parent_node_2)
                        print_to_stderr(
                            "Removing constraint {}".format(
                                parent_node_2.getAttribute("id")
                            )
                        )
        if passed_dom:
            return dom
        utils.replace_cib_configuration(dom)
    return None


# Re-assign any constraints referencing a resource to its parent (a clone
# or master)
def constraint_resource_update(old_id, dom):
    """
    Commandline options: no options
    """
    new_id = None
    clone_ms_parent = utils.dom_get_resource_clone_ms_parent(dom, old_id)
    if clone_ms_parent:
        new_id = clone_ms_parent.getAttribute("id")

    if new_id:
        constraints = dom.getElementsByTagName("rsc_location")
        constraints += dom.getElementsByTagName("rsc_order")
        constraints += dom.getElementsByTagName("rsc_colocation")
        attrs_to_update = ["rsc", "first", "then", "with-rsc"]
        for constraint in constraints:
            for attr in attrs_to_update:
                if constraint.getAttribute(attr) == old_id:
                    constraint.setAttribute(attr, new_id)
    return dom



import json
from unittest import TestCase, mock

from pcs import settings
from pcs.common import reports
from pcs.lib.commands import resource

from pcs_test.tier0.lib.commands.tag import tag_common
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.xml import XmlManipulation


def fixture_primitive_cib_enabled(resource_id="A"):
    return f"""
        <primitive class="ocf" id="{resource_id}" provider="heartbeat" type="Dummy" />
    """


def fixture_primitive_cib_disabled(resource_id="A"):
    return f"""
        <primitive class="ocf" id="{resource_id}" provider="heartbeat" type="Dummy">
            <meta_attributes id="{resource_id}-meta_attributes">
                <nvpair id="{resource_id}-meta_attributes-target-role"
                    name="target-role" value="Stopped" />
            </meta_attributes>
        </primitive>
    """


FIXTURE_MULTIPLE_PRIMITIVES_STATUS = """
    <resources>
        <resource id="A" role="Started" managed="true" />
        <resource id="B" role="Started" managed="true" />
    </resources>
"""

FIXTURE_MULTIPLE_PRIMITIVES_STATUS_UNMANAGED = """
    <resources>
        <resource id="A" role="Started" managed="true" />
        <resource id="B" role="Started" managed="false" />
    </resources>
"""


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_rng/api/api-result.rng")
)
class StopResources(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_primitive(self):
        self.config.runner.cib.load(
            resources=f"""
                <resources>
                    {fixture_primitive_cib_enabled("A")}
                    {fixture_primitive_cib_enabled("B")}
                </resources>
            """
        )
        self.config.runner.pcmk.load_state(
            resources=FIXTURE_MULTIPLE_PRIMITIVES_STATUS
        )
        self.config.env.push_cib(
            resources=f"""
                <resources>
                    {fixture_primitive_cib_disabled("A")}
                    {fixture_primitive_cib_enabled("B")}
                </resources>
            """
        )

        resource.stop(self.env_assist.get_env(), ["A"], set())

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.STOPPING_RESOURCES, resource_id_list=["A"]
                )
            ]
        )

    def test_multiple_primitives(self):
        self.config.runner.cib.load(
            resources=f"""
                <resources>
                    {fixture_primitive_cib_enabled("A")}
                    {fixture_primitive_cib_enabled("B")}
                </resources>
            """
        )
        self.config.runner.pcmk.load_state(
            resources=FIXTURE_MULTIPLE_PRIMITIVES_STATUS
        )
        self.config.env.push_cib(
            resources=f"""
                <resources>
                    {fixture_primitive_cib_disabled("A")}
                    {fixture_primitive_cib_disabled("B")}
                </resources>
            """
        )

        resource.stop(self.env_assist.get_env(), ["A", "B"], set())

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.STOPPING_RESOURCES,
                    resource_id_list=["A", "B"],
                )
            ]
        )

    def test_nonexistent_resource(self):
        self.config.runner.cib.load(
            resources=f"""
                <resources>
                    {fixture_primitive_cib_enabled()}
                </resources>
            """
        )
        self.config.runner.pcmk.load_state()

        self.env_assist.assert_raise_library_error(
            lambda: resource.stop(
                self.env_assist.get_env(), ["nonexistent"], set()
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="nonexistent",
                    expected_types=[
                        "bundle",
                        "clone",
                        "group",
                        "master",
                        "primitive",
                        "tag",
                    ],
                    context_type="cib",
                    context_id="",
                )
            ]
        )

    def test_inner_resources(self):
        self.config.runner.cib.load(
            resources=f"""
                <resources>
                    <clone id="C">
                        <group id="G">
                            {fixture_primitive_cib_enabled("A")}
                            {fixture_primitive_cib_enabled("B")}
                        </group>
                    </clone>
                </resources>
            """
        )

        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <clone id="C" multi_state="true" unique="false" managed="true">
                        <group id="G:0" number_resources="2">
                            <resource id="A" role="Started" managed="true" />
                            <resource id="B" role="Started" managed="true" />
                        </group>
                    </clone>
                </resources>
            """
        )
        self.config.env.push_cib(
            resources=f"""
                <resources>
                    <clone id="C">
                        <group id="G">
                            {fixture_primitive_cib_disabled("A")}
                            {fixture_primitive_cib_disabled("B")}
                        </group>
                    </clone>
                </resources>
            """
        )

        resource.stop(self.env_assist.get_env(), ["C"], set())

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.STOPPING_RESOURCES,
                    resource_id_list=["A", "B"],
                )
            ]
        )

    def test_tag(self):
        tag_fixture = tag_common.fixture_tags_xml([("T", ("A", "B"))])
        self.config.runner.cib.load(
            resources=f"""
                <resources>
                    {fixture_primitive_cib_enabled("A")}
                    {fixture_primitive_cib_enabled("B")}
                </resources>
            """,
            tags=tag_fixture,
        )
        self.config.runner.pcmk.load_state(
            resources=FIXTURE_MULTIPLE_PRIMITIVES_STATUS
        )
        self.config.env.push_cib(
            resources=f"""
                <resources>
                    {fixture_primitive_cib_disabled("A")}
                    {fixture_primitive_cib_disabled("B")}
                </resources>
            """,
            tags=tag_fixture,
        )

        resource.stop(self.env_assist.get_env(), ["T"], set())

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.STOPPING_RESOURCES,
                    resource_id_list=["A", "B"],
                )
            ]
        )

    def test_unmanaged(self):
        self.config.runner.cib.load(
            resources=f"""
                <resources>
                    {fixture_primitive_cib_enabled("A")}
                    {fixture_primitive_cib_enabled("B")}
                </resources>
            """
        )
        self.config.runner.pcmk.load_state(
            resources=FIXTURE_MULTIPLE_PRIMITIVES_STATUS_UNMANAGED
        )

        self.env_assist.assert_raise_library_error(
            lambda: resource.stop(self.env_assist.get_env(), ["A", "B"], set())
        )

        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_IS_UNMANAGED,
                    resource_id="B",
                    force_code=reports.codes.FORCE,
                )
            ]
        )

    def test_unmanaged_force(self):
        self.config.runner.cib.load(
            resources=f"""
                <resources>
                    {fixture_primitive_cib_enabled("A")}
                    {fixture_primitive_cib_enabled("B")}
                </resources>
            """
        )
        self.config.runner.pcmk.load_state(
            resources=FIXTURE_MULTIPLE_PRIMITIVES_STATUS_UNMANAGED
        )
        self.config.env.push_cib(
            resources=f"""
                <resources>
                    {fixture_primitive_cib_disabled("A")}
                    {fixture_primitive_cib_disabled("B")}
                </resources>
            """
        )

        resource.stop(
            self.env_assist.get_env(), ["A", "B"], [reports.codes.FORCE]
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RESOURCE_IS_UNMANAGED, resource_id="B"
                ),
                fixture.info(
                    reports.codes.STOPPING_RESOURCES,
                    resource_id_list=["A", "B"],
                ),
            ]
        )

    def test_bundle_with_clone_works(self):
        self.config.runner.cib.load(
            resources=f"""
                <resources>
                    <bundle id="BUNDLE">
                        <docker image="pcs:test" />
                        {fixture_primitive_cib_enabled("A")}
                    </bundle>
                    <clone>
                        {fixture_primitive_cib_enabled("B")}
                    </clone>
                </resources>
            """
        )
        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <bundle id="BUNDLE" type="podman" managed="true">
                        <replica id="0">
                            <resource id="BUNDLE-ip-192.168.122.250"
                                      resource_agent="ocf:heartbeat:IPaddr2"
                                      role="Stopped"
                            />
                            <resource id="A" role="Stopped" managed="true"/>
                            <resource id="BUNDLE-podman-0"
                                      resource_agent="ocf:heartbeat:podman"
                                      role="Stopped"
                            />
                            <resource id="BUNDLE-0"
                                      resource_agent="ocf:pacemaker:remote"
                                      role="Stopped"
                            />
                        </replica>
                    </bundle>
                    <clone id="C" multi_state="false" unique="false" managed="true">
                        <resource id="B" role="Stopped" managed="true"/>
                    </clone>
                </resources>
            """
        )
        self.config.env.push_cib(
            resources=f"""
                <resources>
                    <bundle id="BUNDLE">
                        <meta_attributes id="BUNDLE-meta_attributes">
                            <nvpair id="BUNDLE-meta_attributes-target-role" name="target-role" value="Stopped"/>
                        </meta_attributes>
                        <docker image="pcs:test" />
                        {fixture_primitive_cib_disabled("A")}
                    </bundle>
                    <clone>
                        {fixture_primitive_cib_enabled("B")}
                    </clone>
                </resources>
            """
        )

        resource.stop(self.env_assist.get_env(), ["BUNDLE"], set())
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.STOPPING_RESOURCES,
                    resource_id_list=["A", "BUNDLE"],
                )
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_rng/api/api-result.rng")
)
class StopStonith(TestCase):
    resources_cib = """
        <resources>
            <primitive id="S1" class="stonith" type="fence_any" />
            <primitive id="S2" class="stonith" type="fence_any" />
            <primitive id="S3" class="stonith" type="fence_kdump" />
        </resources>
    """
    resources_cib_disabled = """
        <resources>
            <primitive id="S1" class="stonith" type="fence_any">
                <meta_attributes id="S1-meta_attributes">
                    <nvpair id="S1-meta_attributes-target-role"
                        name="target-role" value="Stopped"
                    />
                </meta_attributes>
            </primitive>
            <primitive id="S2" class="stonith" type="fence_any">
                <meta_attributes id="S2-meta_attributes">
                    <nvpair id="S2-meta_attributes-target-role"
                        name="target-role" value="Stopped"
                    />
                </meta_attributes>
            </primitive>
            <primitive id="S3" class="stonith" type="fence_kdump" />
        </resources>
    """
    resources_status = """
        <resources>
            <resource id="S1" managed="true" />
            <resource id="S2" managed="true" />
            <resource id="S3" managed="true" />
        </resources>
    """

    def fixture_config_sbd_calls(self, sbd_enabled):
        node_name_list = ["node-1", "node-2"]
        self.config.env.set_known_nodes(node_name_list)
        self.config.corosync_conf.load(node_name_list=node_name_list)
        self.config.http.sbd.check_sbd(
            communication_list=[
                dict(
                    label=node,
                    param_list=[("watchdog", ""), ("device_list", "[]")],
                    output=json.dumps(
                        dict(
                            sbd=dict(
                                installed=True,
                                enabled=sbd_enabled,
                                running=sbd_enabled,
                            )
                        )
                    ),
                )
                for node in node_name_list
            ]
        )

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_useful_enabled_stonith_left(self):
        resources_cib_disabled = """
            <resources>
                <primitive id="S1" class="stonith" type="fence_any" />
                <primitive id="S2" class="stonith" type="fence_any">
                    <meta_attributes id="S2-meta_attributes">
                        <nvpair id="S2-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                </primitive>
                <primitive id="S3" class="stonith" type="fence_kdump" />
            </resources>
        """
        self.config.runner.cib.load(resources=self.resources_cib)
        self.config.runner.pcmk.load_state(resources=self.resources_status)
        self.config.env.push_cib(resources=resources_cib_disabled)

        resource.stop(self.env_assist.get_env(), ["S2"])
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.STOPPING_RESOURCES, resource_id_list=["S2"]
                )
            ]
        )

    def test_no_stonith_left_sbd_enabled(self):
        self.config.runner.cib.load(resources=self.resources_cib)
        self.fixture_config_sbd_calls(True)
        self.config.runner.pcmk.load_state(resources=self.resources_status)
        self.config.env.push_cib(resources=self.resources_cib_disabled)

        resource.stop(self.env_assist.get_env(), ["S1", "S2"])
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.STOPPING_RESOURCES,
                    resource_id_list=["S1", "S2"],
                )
            ]
        )

    def test_no_stonith_left_not_live(self):
        tmp_file = "/fake/tmp_file"
        cmd_env = dict(CIB_file=tmp_file)
        cib_xml_man = XmlManipulation.from_file(rc("cib-empty.xml"))
        cib_xml_man.append_to_first_tag_name("resources", self.resources_cib)
        self.config.env.set_cib_data(str(cib_xml_man), cib_tempfile=tmp_file)
        self.config.runner.cib.load(resources=self.resources_cib, env=cmd_env)
        self.config.runner.pcmk.load_state(
            resources=self.resources_status, env=cmd_env
        )
        # doesn't call other nodes to check sbd status

        self.env_assist.assert_raise_library_error(
            lambda: resource.stop(self.env_assist.get_env(), ["S1", "S2"])
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.NO_STONITH_MEANS_WOULD_BE_LEFT,
                    force_code=reports.codes.FORCE,
                )
            ]
        )

    def test_no_stonith_left_sbd_disabled(self):
        self.config.runner.cib.load(resources=self.resources_cib)
        self.fixture_config_sbd_calls(False)
        self.config.runner.pcmk.load_state(resources=self.resources_status)

        self.env_assist.assert_raise_library_error(
            lambda: resource.stop(self.env_assist.get_env(), ["S1", "S2"])
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.NO_STONITH_MEANS_WOULD_BE_LEFT,
                    force_code=reports.codes.FORCE,
                )
            ]
        )

    def test_no_stonith_left_sbd_disabled_forced(self):
        self.config.runner.cib.load(resources=self.resources_cib)
        self.fixture_config_sbd_calls(False)
        self.config.runner.pcmk.load_state(resources=self.resources_status)
        self.config.env.push_cib(resources=self.resources_cib_disabled)

        resource.stop(
            self.env_assist.get_env(),
            ["S1", "S2"],
            force_flags={reports.codes.FORCE},
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.NO_STONITH_MEANS_WOULD_BE_LEFT,
                ),
                fixture.info(
                    reports.codes.STOPPING_RESOURCES,
                    resource_id_list=["S1", "S2"],
                ),
            ]
        )

    def test_no_useful_enabled_stonith_removed(self):
        resources_cib = """
            <resources>
                <primitive id="S1" class="stonith" type="fence_any">
                    <meta_attributes id="S1-meta_attributes">
                        <nvpair id="S1-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                </primitive>
                <primitive id="S3" class="stonith" type="fence_kdump" />
            </resources>
        """
        resources_cib_disabled = """
            <resources>
                <primitive id="S1" class="stonith" type="fence_any">
                    <meta_attributes id="S1-meta_attributes">
                        <nvpair id="S1-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                </primitive>
                <primitive id="S3" class="stonith" type="fence_kdump">
                    <meta_attributes id="S3-meta_attributes">
                        <nvpair id="S3-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                </primitive>
            </resources>
        """
        self.config.runner.cib.load(resources=resources_cib)
        self.fixture_config_sbd_calls(False)
        self.config.runner.pcmk.load_state(resources=self.resources_status)
        self.config.env.push_cib(resources=resources_cib_disabled)

        resource.stop(self.env_assist.get_env(), ["S1", "S3"])
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.STOPPING_RESOURCES,
                    resource_id_list=["S1", "S3"],
                )
            ]
        )



import uuid
from dataclasses import astuple, dataclass
from typing import Generator, MutableSet, Optional, TypeVar, Union

from lxml import etree
from lxml.etree import _Element

from pcs.common.types import StringCollection
from pcs.common.validate import is_integer

T = TypeVar("T", bound=type)


def bin_to_str(binary: bytes) -> str:
    return "".join(map(chr, binary))


def get_all_subclasses(cls: T) -> MutableSet[T]:
    subclasses = set(cls.__subclasses__())
    return subclasses.union(
        {s for c in subclasses for s in get_all_subclasses(c)}
    )


def get_unique_uuid(already_used: StringCollection) -> str:
    is_duplicate = True
    while is_duplicate:
        candidate = str(uuid.uuid4())
        is_duplicate = candidate in already_used
    return candidate


def format_os_error(e: OSError) -> str:
    if e.filename:
        return f"{e.strerror}: '{e.filename}'"
    if e.strerror:
        return e.strerror
    return (f"{e.__class__.__name__} {e}").strip()


def xml_fromstring(xml: str) -> _Element:
    # If the xml contains encoding declaration such as:
    # <?xml version="1.0" encoding="UTF-8"?>
    # we get an exception in python3:
    # ValueError: Unicode strings with encoding declaration are not supported.
    # Please use bytes input or XML fragments without declaration.
    # So we encode the string to bytes.
    return etree.fromstring(
        xml.encode("utf-8"),
        # it raises on a huge xml without the flag huge_tree=True
        # see https://bugzilla.redhat.com/show_bug.cgi?id=1506864
        etree.XMLParser(huge_tree=True),
    )


def timeout_to_seconds(timeout: Union[int, str]) -> Optional[int]:
    """
    Transform pacemaker style timeout to number of seconds. If `timeout` is not
    a valid timeout, `None` is returned.

    timeout -- timeout string
    """
    try:
        candidate = int(timeout)
        if candidate >= 0:
            return candidate
        return None
    except ValueError:
        pass
    # Now we know the timeout is not an integer nor an integer string.
    # Let's make sure mypy knows the timeout is a string as well.
    timeout = str(timeout)
    suffix_multiplier = {
        "s": 1,
        "sec": 1,
        "m": 60,
        "min": 60,
        "h": 3600,
        "hr": 3600,
    }
    for suffix, multiplier in suffix_multiplier.items():
        if timeout.endswith(suffix):
            candidate2 = timeout[: -len(suffix)]
            if is_integer(candidate2, at_least=0):
                return int(candidate2) * multiplier
    return None


@dataclass(frozen=True)
class Version:
    major: int
    minor: Optional[int] = None
    revision: Optional[int] = None

    @property
    def as_full_tuple(self) -> tuple[int, int, int]:
        return (
            self.major,
            self.minor if self.minor is not None else 0,
            self.revision if self.revision is not None else 0,
        )

    def normalize(self) -> "Version":
        return self.__class__(*self.as_full_tuple)

    def __iter__(self) -> Generator[Optional[int], None, None]:
        yield from astuple(self)

    def __getitem__(self, index: int) -> Optional[int]:
        return astuple(self)[index]

    def __str__(self) -> str:
        return ".".join([str(x) for x in self if x is not None])

    def __lt__(self, other: "Version") -> bool:
        return self.as_full_tuple < other.as_full_tuple

    def __le__(self, other: "Version") -> bool:
        return self.as_full_tuple <= other.as_full_tuple

    def __hash__(self) -> int:
        # https://docs.astral.sh/ruff/rules/eq-without-hash/
        # used self.as_full_tuple because __eq__ and __hash__ should be in sync,
        # objects which compare equal must have the same hash value
        return hash(self.as_full_tuple)

    # See, https://stackoverflow.com/questions/37557411/why-does-defining-the-argument-types-for-eq-throw-a-mypy-type-error
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Version):
            return NotImplemented
        return self.as_full_tuple == other.as_full_tuple

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Version):
            return NotImplemented
        return self.as_full_tuple != other.as_full_tuple

    def __gt__(self, other: "Version") -> bool:
        return self.as_full_tuple > other.as_full_tuple

    def __ge__(self, other: "Version") -> bool:
        return self.as_full_tuple >= other.as_full_tuple



from unittest import TestCase

from lxml import etree

from pcs.lib.booth import cib as cib_commands


class GetTicketNamesBase:
    def setUp(self):
        self.cib = etree.fromstring(
            """
            <cib>
                <configuration/>
                <status>
                    <tickets>
                        <ticket_state id="T1" granted="false" booth-cfg-name="booth"/>
                        <ticket_state id="T2" granted="false" booth-cfg-name="booth"/>
                        <ticket_state id="T3" granted="false" booth-cfg-name="custom_booth"/>
                        <ticket_state id="T-self-managed" granted="false"/>
                    </tickets>
                </status>
            </cib>
        """
        )


class GetBoothTicketNames(GetTicketNamesBase, TestCase):
    def test_default(self):
        self.assertEqual(
            ["T1", "T2"], cib_commands.get_booth_ticket_names(self.cib)
        )

    def test_custom_booth_name(self):
        self.assertEqual(
            ["T3"],
            cib_commands.get_booth_ticket_names(self.cib, "custom_booth"),
        )

    def test_no_tickets(self):
        self.assertEqual(
            [],
            cib_commands.get_booth_ticket_names(self.cib, "nonexistent_booth"),
        )


class GetTicketNames(GetTicketNamesBase, TestCase):
    def test_success(self):
        self.assertEqual(
            ["T1", "T2", "T3", "T-self-managed"],
            cib_commands.get_ticket_names(self.cib),
        )
